/******************************************************************************
** blastjob.cpp
** Implements the BlastJob class - manages resource allocation for BLAST jobs
** $Id: blastjob.cpp,v 1.18 2005/06/20 03:58:52 koadman Exp $
** $Revision: 1.18 $
** 
** This file is a part of mpiBLAST and is copyright (c)
** by Aaron Darling of Los Alamos National Laboratory for the Regents
** of the University of California.
** 
** License:
** 
** Copyright 2002-2005. The Regents of the University of California.
** 
** This material was produced under U.S. Government contract
** W-7405-ENG-36 for Los Alamos National Laboratory, which is operated
** by the University of California for the U.S. Department of Energy.
** The Government is granted for itself and other acting on its behalf
** a paid-up, non-exclusive, irrevocable worldwide license in the
** material to reproduce, prepare derivative works, and perform
** publicly and display publicly. Beginning five (5) years after
** November 6, 2002, subject to additional five-year worldwide
** renewals, the Government is granted for itself and others acting on
** its behalf a paid-up, nonexclusive, irrevocable worldwide license
** in this material to reproduce, prepare derivative works, distribute
** copies to the public, perform publicly and display publicly, and to
** permit others to do so. NEITHER THE UNITED STATES NOR THE UNITED
** STATES DEPARTMENT OF ENERGY, NOR THE UNIVERSITY OF CALIFORNIA, NOR
** ANY OF THEIR EMPLOYEES, MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR
** ASSUMES ANY LEGAL LIABLITY OR RESPONSIBILITY FOR THE ACCURACY,
** COMPLETENESS, OR USEFULNESS OF ANY INFORMATION, APPARATUS, PRODUCT,
** OR PROCESS DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE
** PRIVATELY OWNED RIGHTS.
** 
** Additionally, this program is free software; you can distribute it
** and/or modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2 of
** the License, or any later version. Accordingly, this program is
** distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or
** FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
** License for more details. You should have received a copy of the
** GNU General Public License along with this program; if not, write
** to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
** Boston, MA 02111-1307, USA.
** 
** ------------------------------------------------------------
** 
** See the file 'COPYING' for the details of the GNU General Public License
** 
******************************************************************************/

#include "blastjob.hpp"
#include "mpiblast.hpp"
#include <algorithm>
using namespace std;

BlastJob::BlastJob( int frag_count, FragmentDistribution& frag_dist ) :
fragment_count( frag_count ),
frag_distribution( frag_dist )
{	
	// init copy_fragments and unassigned_fragments
	uint fragmentI;
	for( fragmentI = 0; fragmentI < (uint)fragment_count; fragmentI++ ){
		unassigned_fragments.insert( fragmentI );
	}
}

BlastJob::BlastJob( const BlastJob& bj ) :
fragment_count( bj.fragment_count ),
frag_distribution( bj.frag_distribution ),
unassigned_fragments( bj.unassigned_fragments )
{
}

BlastJob& BlastJob::operator=( const BlastJob& bj ){
	BlastJob( bj.fragment_count, bj.frag_distribution );
	unassigned_fragments = bj.unassigned_fragments;
	return *this;
}

void BlastJob::getAssignment( int nodeI, int& operation, int& fragment_id ){
	if( debug_msg )
		LOG_MSG << "Computing next assignment for node " << nodeI << endl;

	// first try assigning the least distributed unassigned fragment that
	// already exists on nodeI
	// intersect unassigned with the set of frags on nodeI
	set< int > nodeI_unassigned;
	insert_iterator< set< int > > node_ins(nodeI_unassigned, nodeI_unassigned.begin());
	set_intersection( frag_distribution.getNodeFrags( nodeI ).begin(), frag_distribution.getNodeFrags( nodeI ).end(),
			unassigned_fragments.begin(), unassigned_fragments.end(), node_ins );
	
	// print some debugging info
	if( debug_msg ){
		LOG_MSG << "Node " << nodeI << " fragments: ";
		frag_distribution.printNodeFrags( nodeI );
		CONT_LOG_MSG << endl;
		
		LOG_MSG << "Unassigned fragments: ";
		set< int >::const_iterator iter = unassigned_fragments.begin();
		for( ; iter != unassigned_fragments.end(); iter++ ){
			CONT_LOG_MSG << " " << *iter;
		}
		CONT_LOG_MSG << endl;

		LOG_MSG << "Intersection fragments: ";
		iter = nodeI_unassigned.begin();
		for( ; iter != nodeI_unassigned.end(); iter++ ){
			CONT_LOG_MSG << " " << *iter;
		}
		CONT_LOG_MSG << endl;
	}


	// get the least distributed
	int search_fragI = frag_distribution.getLeastDistributed( nodeI_unassigned );	
	if( debug_msg )
		LOG_MSG << "search_fragI is " << search_fragI << endl;

	if( search_fragI != -1 ){
		unassigned_fragments.erase( search_fragI );
		operation = SEARCH_FRAGMENT;
		fragment_id = search_fragI;
		return;
	}
	
	// if -1 was returned (the intersected set was empty) then we need to copy a frag.
	// find the least distributed unassigned fragment to copy
	// take the set difference of unassigned and copy_fragments
	
	// WARNING!  This code only permits one node to copy a particular DB fragment at a time
	// This behavior is fine so long as query segmentation is not implemented
	set< int > uncopy;
	insert_iterator< set< int > > unass_ins(uncopy, uncopy.begin());
	set_difference( unassigned_fragments.begin(), unassigned_fragments.end(),
			frag_distribution.getCopyFrags().begin(), frag_distribution.getCopyFrags().end(),
			unass_ins );

	int copy_fragI = frag_distribution.getLeastDistributed( uncopy );
	
	if( copy_fragI != -1 ){
		operation = COPY_FRAGMENT;
		fragment_id = copy_fragI;
		// JA -- Once a node copies a fragment, it immediately starts searching it.
		//       This bypasses the Worker asking for an assignment, which we know will be SEARCH_FRAGMENT
//		unassigned_fragments.erase( fragment_id );
		frag_distribution.copyFragment( copy_fragI );
		return;
	}
	
	// if all of the fragments waiting to be assigned are already being copied
	// then this worker can't do any work on this query and should move on to the next
	// query.  return a state of DO_NOTHING
	fragment_id = 0;
	if( frag_distribution.getCopyFrags().size() > 0 )
		operation = DO_NOTHING;
	else
		// if -1 was returned (the intersected set was empty) then the search must be complete
		operation = SEARCH_COMPLETE;

}

int PrecopySchedulerPolicy::db_copy_number = 1;

PrecopySchedulerPolicy::PrecopySchedulerPolicy( int fragment_count, FragmentDistribution& frag_distribution )
: BlastJob( fragment_count, frag_distribution )
{
}

PrecopySchedulerPolicy::PrecopySchedulerPolicy( const PrecopySchedulerPolicy& bj )
: BlastJob( bj ) {}

PrecopySchedulerPolicy& PrecopySchedulerPolicy::operator=( const PrecopySchedulerPolicy& bj )
{
	BlastJob::operator =( bj );
	return *this;
}

void PrecopySchedulerPolicy::getAssignment( int nodeI, int& operation, int& fragment_id )
{
	set< int > all_frags;
	for( uint fragI = 0; fragI < fragment_count; fragI++ )
		all_frags.insert( fragI );
	set< int > cur_frags = all_frags;
	const set< int >& cf = frag_distribution.getCopyFrags();
	// if this worker has no fragment, tell it to copy the least distributed
	// fragment
	if( frag_distribution.getNodeFrags( nodeI ).size() == 0 ){
		int copy_fragI = -1;
		int least_count = -1;
		while( copy_fragI == -1 ){
			int least_frag = frag_distribution.getLeastDistributed( cur_frags );
			if( least_count == -1 )
				least_count = frag_distribution.getFragNodes(least_frag).size();
			if( least_count < frag_distribution.getFragNodes(least_frag).size() )
				break;	// failed to make an assignment
			if( cf.find( least_frag ) != cf.end() ){
					cur_frags.erase( least_frag );
					continue;	// this one is being copied already.  keep looking.
			}
			copy_fragI = least_frag;
		}
		if( copy_fragI != -1 ){
			operation = COPY_FRAGMENT;
			fragment_id = copy_fragI;
			frag_distribution.copyFragment( copy_fragI );
			return;
		}
		// if the minimally distributed fragments are already being copied
		// tell this worker to do nothing for the time being
		fragment_id = 0;
		if( frag_distribution.getCopyFrags().size() > 0 )
			operation = DO_NOTHING;
		return;
	}

	// ensure the fragment distribution is complete
	// if an undistributed fragment is not already being copied
	// then tell this worker to copy it!

	cur_frags = all_frags;
	int copy_fragI = frag_distribution.getLeastDistributed( cur_frags );
	while( 
		frag_distribution.getFragNodes( copy_fragI ).size() < db_copy_number &&
		(cf.find( copy_fragI ) != cf.end() || 
		frag_distribution.getFragNodes( copy_fragI ).find( nodeI ) !=
		frag_distribution.getFragNodes( copy_fragI ).end() )
		)
	{
			cur_frags.erase( copy_fragI );
			if( cur_frags.size() == 0 ){
				copy_fragI = -1;
				break;
			}
			copy_fragI = frag_distribution.getLeastDistributed( cur_frags );
			continue;	// this one is being copied already.  keep looking.
	}
	if( copy_fragI != -1 && frag_distribution.getFragNodes( copy_fragI ).size() < db_copy_number ){
		// this guy needs to get copied
		operation = COPY_FRAGMENT;
		fragment_id = copy_fragI;
		frag_distribution.copyFragment( copy_fragI );
		return;
	}

	// if we get here then the DB has been minimally distributed
	// from here on out, assign the least distributed fragment to get searched

	// intersect unassigned with the set of frags on nodeI
	set< int > nodeI_unassigned;
	insert_iterator< set< int > > node_ins(nodeI_unassigned, nodeI_unassigned.begin());
	set_intersection( frag_distribution.getNodeFrags( nodeI ).begin(), frag_distribution.getNodeFrags( nodeI ).end(),
			unassigned_fragments.begin(), unassigned_fragments.end(), node_ins );

	// get the least distributed
	int search_fragI = frag_distribution.getLeastDistributed( nodeI_unassigned );	

	if( search_fragI != -1 ){
		unassigned_fragments.erase( search_fragI );
		operation = SEARCH_FRAGMENT;
		fragment_id = search_fragI;
		return;
	}
	
	fragment_id = 0;
	if( unassigned_fragments.size() == 0 )
		operation = SEARCH_COMPLETE;
	else
		operation = DO_NOTHING;
	return;
}

FragmentDistribution::FragmentDistribution( uint fragment_count, uint node_count ) {

	set< int > frag_set;
	fragment_nodes = vector< set< int > >( fragment_count );
	node_fragments = vector< set< int > >( node_count );
	
}

FragmentDistribution::FragmentDistribution( const FragmentDistribution& fd ) {
	*this = fd;
}

FragmentDistribution& FragmentDistribution::operator=( const FragmentDistribution& fd ) {
	node_fragments = fd.node_fragments;
	fragment_nodes = fd.fragment_nodes;
	copy_fragments = fd.copy_fragments;
	return *this;
}

void FragmentDistribution::update( int nodeI, const std::set< int >& fragment_list ){
	node_fragments[ nodeI ] = fragment_list;
	
	// remove this node from all of the fragment lists
	for( int fragI = 0; fragI < fragment_nodes.size(); fragI++ ){
		fragment_nodes[ fragI ].erase( nodeI );
	}
	
	// add this node to all of the correct fragment node lists
	set< int >::const_iterator frag_iter = fragment_list.begin();
	for( ; frag_iter != fragment_list.end(); frag_iter++ ){
		fragment_nodes[ *frag_iter ].insert( nodeI );
	}
}

void FragmentDistribution::printNodeFrags( int nodeI ) const {
	set< int >::const_iterator iter = node_fragments[ nodeI ].begin();
	for( ; iter != node_fragments[ nodeI ].end(); iter++ ){
		CONT_LOG_MSG << " " << *iter;
	}
}

const set<int>& FragmentDistribution::getNodeFrags( int nodeI ) const {
	return node_fragments[ nodeI ];
}

const set<int>& FragmentDistribution::getFragNodes( int fragmentI ) const {
	return fragment_nodes[ fragmentI ];
}

const set<int>& FragmentDistribution::getCopyFrags() const {
	return copy_fragments;
}

void FragmentDistribution::copyFragment( int fragmentI ) {
	copy_fragments.insert( fragmentI );
}

void FragmentDistribution::copyCompleted( int nodeI, int fragment_id ) {
	fragment_nodes[ fragment_id ].insert( nodeI );
	copy_fragments.erase( fragment_id );
	node_fragments[ nodeI ].insert( fragment_id );
	
	if(debug_msg){
		LOG_MSG << "node " << nodeI << " copied fragment " << fragment_id << endl;
	}
}


int FragmentDistribution::getLeastDistributed( const set< int >& possible_fragments ) const {
	// find the lowest replication fragment in possible_fragments

	int best_fragment = -1;
	int lowest_replication = -1;
	set< int >::const_iterator frag_iter = possible_fragments.begin();
	
	for(; frag_iter != possible_fragments.end(); frag_iter++ ){
		int fragmentI = *frag_iter;
		if( (fragment_nodes[ fragmentI ].size() < lowest_replication) || 
		   (lowest_replication == -1) ){
			best_fragment = fragmentI;
			lowest_replication = fragment_nodes[ fragmentI ].size();
		}
	}
	return best_fragment;
}


