/******************************************************************************
** blastjob.hpp
** Provides the programming interface for the BlastJob class
** $Id: blastjob.hpp,v 1.17 2005/06/20 03:58:52 koadman Exp $
** $Revision: 1.17 $
** 
** This file is a part of mpiBLAST and is copyright (c)
** by Aaron Darling of Los Alamos National Laboratory for the Regents
** of the University of California.
** 
** License:
** 
** Copyright 2002-2005. The Regents of the University of California.
** 
** This material was produced under U.S. Government contract
** W-7405-ENG-36 for Los Alamos National Laboratory, which is operated
** by the University of California for the U.S. Department of Energy.
** The Government is granted for itself and other acting on its behalf
** a paid-up, non-exclusive, irrevocable worldwide license in the
** material to reproduce, prepare derivative works, and perform
** publicly and display publicly. Beginning five (5) years after
** November 6, 2002, subject to additional five-year worldwide
** renewals, the Government is granted for itself and others acting on
** its behalf a paid-up, nonexclusive, irrevocable worldwide license
** in this material to reproduce, prepare derivative works, distribute
** copies to the public, perform publicly and display publicly, and to
** permit others to do so. NEITHER THE UNITED STATES NOR THE UNITED
** STATES DEPARTMENT OF ENERGY, NOR THE UNIVERSITY OF CALIFORNIA, NOR
** ANY OF THEIR EMPLOYEES, MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR
** ASSUMES ANY LEGAL LIABLITY OR RESPONSIBILITY FOR THE ACCURACY,
** COMPLETENESS, OR USEFULNESS OF ANY INFORMATION, APPARATUS, PRODUCT,
** OR PROCESS DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE
** PRIVATELY OWNED RIGHTS.
** 
** Additionally, this program is free software; you can distribute it
** and/or modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2 of
** the License, or any later version. Accordingly, this program is
** distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or
** FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
** License for more details. You should have received a copy of the
** GNU General Public License along with this program; if not, write
** to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
** Boston, MA 02111-1307, USA.
** 
** ------------------------------------------------------------
** 
** See the file 'COPYING' for the details of the GNU General Public License
** 
******************************************************************************/

#ifndef __BlastJob_hpp__
#define __BlastJob_hpp__

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "mpi.h"

#include <string>
#include <set>
#include <vector>

#include "mpiblast_types.h"

/**
 * Tracks which nodes have which fragments.  A single instance of this
 * object should be created and used to track the DB distribution 
 * among all participating workers
 */
class FragmentDistribution {
public:
	FragmentDistribution(){};
	/**
	 * Construct an empty fragment distribution with a given number of nodes and fragments
	 */
	FragmentDistribution( uint fragment_count, uint node_count );
	FragmentDistribution( const FragmentDistribution& fd );
	FragmentDistribution& operator=( const FragmentDistribution& fd );
	
	/** Sets the fragments on a node's local storage */
	void update( int nodeI, const std::set<int>& fragment_list );
	
	/** Prints the node_fragments to cout */
	void printNodeFrags( int nodeI ) const;
	
	/** Returns the set of fragments on a given node */
	const std::set<int>& getNodeFrags( int nodeI ) const;

	/** Returns the set of nodes that have a given fragment */
	const std::set<int>& getFragNodes( int fragmentI ) const;
	
	/** Return the set of fragments currently being copied */
	const std::set<int>& getCopyFrags() const;

	/** Register that a fragment is being copied */
	void copyFragment( int fragmentI );

	/** Unregister a fragment copy that has completed, add it to the node's list */
	void copyCompleted( int nodeI, int fragment_id );

	/**
	 * Given a set of fragments, find the least distributed fragment
	 */
	int getLeastDistributed( const std::set< int >& possible_fragments ) const;
protected:
	std::vector< std::set< int > > node_fragments;	/**< The set of fragments residing on each node -- index[node][fragment] */
	std::vector< std::set< int > > fragment_nodes;	/**< The set of nodes each fragment resides on -- index[fragment][node] */
	std::set< int > copy_fragments; /**< Fragments that are currently being copied somewhere */
};


/**
 * Manages each Blast job by allocating work to worker nodes according to a policy
 * A "blast job" can be a single query or a set of queries.
 * Each BlastJob contains a reference to a single instance of the FragmentDistribution
 * object for all participating workers.  The BlastJob class can then make fragment
 * search assignments to workers based on knowledge of the database fragment 
 * distribution.
 * 
 */
class BlastJob{
	public:
		BlastJob( int fragment_count, FragmentDistribution& frag_distribution );
		BlastJob( const BlastJob& bj );
		BlastJob& operator=( const BlastJob& bj );
		
		/**
		 * Calculates the best assignment for a node based on the current
		 * distribution of database fragments
		 */
		virtual void getAssignment( int nodeI, int& operation, int& fragment_id );

	protected:
		int fragment_count;	/**< The total number of fragments for this database */
		std::set< int > unassigned_fragments;	/**< Fragments that haven't been assigned yet */
		FragmentDistribution& frag_distribution;	/**< Tracks which nodes have which DB fragments */
};

/** 
 * A scheduler policy that precopies a minimal database 
 * distribution prior to making search assignments
 */
class PrecopySchedulerPolicy : public BlastJob
{
public:
	PrecopySchedulerPolicy( int fragment_count, FragmentDistribution& frag_distribution );
	PrecopySchedulerPolicy( const PrecopySchedulerPolicy& bj );
	PrecopySchedulerPolicy& operator=( const PrecopySchedulerPolicy& bj );
	virtual void getAssignment( int nodeI, int& operation, int& fragment_id );
	static int db_copy_number;	/**< The number of complete copies of the database that will be created across cluster nodes */
};

#endif	//__BlastJob_h__
