/******************************************************************************
** file_util.hpp
** Provides a cross-platform interface to the filesystem.
** $Id: file_util.hpp,v 1.10 2005/06/20 03:58:52 koadman Exp $
** $Revision: 1.10 $
** 
** This file is a part of mpiBLAST and is copyright (c)
** by Aaron Darling of Los Alamos National Laboratory for the Regents
** of the University of California.
** 
** License:
** 
** Copyright 2002-2005. The Regents of the University of California.
** 
** This material was produced under U.S. Government contract
** W-7405-ENG-36 for Los Alamos National Laboratory, which is operated
** by the University of California for the U.S. Department of Energy.
** The Government is granted for itself and other acting on its behalf
** a paid-up, non-exclusive, irrevocable worldwide license in the
** material to reproduce, prepare derivative works, and perform
** publicly and display publicly. Beginning five (5) years after
** November 6, 2002, subject to additional five-year worldwide
** renewals, the Government is granted for itself and others acting on
** its behalf a paid-up, nonexclusive, irrevocable worldwide license
** in this material to reproduce, prepare derivative works, distribute
** copies to the public, perform publicly and display publicly, and to
** permit others to do so. NEITHER THE UNITED STATES NOR THE UNITED
** STATES DEPARTMENT OF ENERGY, NOR THE UNIVERSITY OF CALIFORNIA, NOR
** ANY OF THEIR EMPLOYEES, MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR
** ASSUMES ANY LEGAL LIABLITY OR RESPONSIBILITY FOR THE ACCURACY,
** COMPLETENESS, OR USEFULNESS OF ANY INFORMATION, APPARATUS, PRODUCT,
** OR PROCESS DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE
** PRIVATELY OWNED RIGHTS.
** 
** Additionally, this program is free software; you can distribute it
** and/or modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2 of
** the License, or any later version. Accordingly, this program is
** distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or
** FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
** License for more details. You should have received a copy of the
** GNU General Public License along with this program; if not, write
** to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
** Boston, MA 02111-1307, USA.
** 
** ------------------------------------------------------------
** 
** See the file 'COPYING' for the details of the GNU General Public License
** 
******************************************************************************/

#ifndef __MPIBLAST_FILE_UTIL
#define __MPIBLAST_FILE_UTIL

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string>
#include <fstream>
#include <vector>
#include "mpiblast_types.h"

// copy styles 
#define COPY_VIA_NONE 0 /* don't copy db fragments. useful for PVFS,GPFS,PFS */
#define COPY_VIA_CP   1 /* copy db fragments using cp */
#define COPY_VIA_RCP  2 /* copy db fragments using rcp */
#define COPY_VIA_SCP  3 /* copy db fragments using scp */
#define COPY_VIA_MPI  4 /* copy db fragments using MPI */

#define COPY_VIA_XCP  1 /* copyFile calls that use this need to be changed. this is a hack for right now */



#ifdef WIN32
const std::string PATH_SEPARATOR = "\\";
#else
#include "unistd.h"
#include <sys/types.h>
#include <sys/param.h>
const std::string PATH_SEPARATOR = "/";
#if defined(__MACH__)
# include <sys/mount.h>
# define STATFS statfs
#else
# include <sys/statvfs.h>
# define STATFS statvfs
#endif

#endif

/** Extension for a database's local fragment list file */
#define	FRAG_LIST_EXTENSION 	".mbf"

bool doesFileExist(const std::string& file_name );
std::string getPath(const std::string& file_path );
std::string getFilename(const std::string& file_path);
void getTempFileName(std::string& tempname);
int moveFile(const std::string& src, const std::string& dest);
int copyFile(const std::string& src, const std::string& dest, const int fileType);
int removeFile(const std::string& filename, bool verbose = false);

bool removeDirectory( const std::string& dir );

/** get the size in bytes of a particular file */
uint64 statFileSize( const char* path );

/** get the modification time of a file */
time_t statFileMTime( const char* path );

/** 
 * Check that a directory exists and if not try to create it
 * @return true if the directory exists or was created, false otherwise
 */
bool checkCreateDir( const std::string& path );

/** 
 * Check that we have write permissions to a directory
 * @return true if we have write permission, false otherwise
 */
bool checkDirWritePerms( const std::string& path );

/**
 * Set an environment variable in a cross platform manner
 */
void setEnvironmentVariable( const char* var );

/**
 * Check whether a path specifies an rcp or scp host
 * If a path contains a ":" character and the ":" is not part of a
 * Windows drive identifier then return true.  fixme: how can
 * we distinguish Mac OS X paths with : from remote hosts?
 */
bool isRemotePath( const std::string& path );

/**
 * Register a file name to be deleted before the process exits
 * When passed an empty string, it does not add to the list of files to delete
 * @param fname	The name of a file to delete, empty strings are ignored
 * @return A vector of file names registered for deletion
 */
std::vector< std::string >& registerFileToDelete( std::string fname = "" );

/* Returns free space in KB for a filesystem in a given path */
long freeSpaceKB(const std::string& path);

/**
 * Deletes files associated with mpiBLAST from a given directory
 * Recursively deletes temp directories associated with mpiBLAST
 * this function needs to execute the equivalent of
 * rm -fr /$TMPDIR/mpiblast_tmpXXXXXX
 * rm -fr /$TMPDIR/*.[np]?? /$TMPDIR/*.mbf /$TMPDIR/*.dbs
 * rm -fr /$TMPDIR/*.[np]??XXXXXX /$TMPDIR/*.mbfXXXXXX /$TMPDIR/*.dbsXXXXXX
 */
void deleteMpiBlastFiles( const std::string& directory );

#endif // __MPIBLAST_FILE_UTIL
