/******************************************************************************
** fragment_list.cpp
** Parses and provides an API to the fragment list on worker nodes
** $Id: fragment_list.cpp,v 1.21 2005/06/20 03:58:52 koadman Exp $
** $Revision: 1.21 $
** 
** This file is a part of mpiBLAST and is copyright (c)
** by Aaron Darling of Los Alamos National Laboratory for the Regents
** of the University of California.
** 
** License:
** 
** Copyright 2002-2005. The Regents of the University of California.
** 
** This material was produced under U.S. Government contract
** W-7405-ENG-36 for Los Alamos National Laboratory, which is operated
** by the University of California for the U.S. Department of Energy.
** The Government is granted for itself and other acting on its behalf
** a paid-up, non-exclusive, irrevocable worldwide license in the
** material to reproduce, prepare derivative works, and perform
** publicly and display publicly. Beginning five (5) years after
** November 6, 2002, subject to additional five-year worldwide
** renewals, the Government is granted for itself and others acting on
** its behalf a paid-up, nonexclusive, irrevocable worldwide license
** in this material to reproduce, prepare derivative works, distribute
** copies to the public, perform publicly and display publicly, and to
** permit others to do so. NEITHER THE UNITED STATES NOR THE UNITED
** STATES DEPARTMENT OF ENERGY, NOR THE UNIVERSITY OF CALIFORNIA, NOR
** ANY OF THEIR EMPLOYEES, MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR
** ASSUMES ANY LEGAL LIABLITY OR RESPONSIBILITY FOR THE ACCURACY,
** COMPLETENESS, OR USEFULNESS OF ANY INFORMATION, APPARATUS, PRODUCT,
** OR PROCESS DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE
** PRIVATELY OWNED RIGHTS.
** 
** Additionally, this program is free software; you can distribute it
** and/or modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2 of
** the License, or any later version. Accordingly, this program is
** distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or
** FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
** License for more details. You should have received a copy of the
** GNU General Public License along with this program; if not, write
** to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
** Boston, MA 02111-1307, USA.
** 
** ------------------------------------------------------------
** 
** See the file 'COPYING' for the details of the GNU General Public License
** 
******************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "mpi.h"

#include "fragment_list.hpp"
#include "mpiblast.hpp"
#include "mpiblast_util.hpp"

extern "C" {
#include "blast_hooks.h"
#include "readdb.h"
}

using namespace std;

bool lock_mbf = false;

/*
 * This function uses platform-specific I/O since accesses to the
 * fragment list file must be synchronous (we lock the file)
 */
FragmentListFile::FragmentListFile( const string& filename, const MpiBlastConfig& config, const std::string& database_name, const std::string& db_type ) : 
config( config ), 
database_name( database_name ),
db_type( db_type )
{
  frag_filename = filename;
#ifdef WIN32
	HANDLE fragfile = openAndReadWin32();
	if( fragfile == INVALID_HANDLE_VALUE )
		throw "Unable to access fragment list file";
	CloseHandle( fragfile );
#else
	int fragfile = openAndReadPOSIX();
	if( fragfile == -1 )
		throw "Unable to access fragment list file";
	close( fragfile );
#endif
}

FragmentListFile& FragmentListFile::operator=( const FragmentListFile& mbc ){
  fragment_list = mbc.fragment_list;
  fragment_set = mbc.fragment_set;
  frag_filename = mbc.frag_filename;

	config = mbc.config;
	database_name = mbc.database_name;
	db_type = mbc.db_type ;

	return *this;
}

/**
 * Populate the fragment list from an input stream containing a fragment list
 */
void FragmentListFile::parseStream( istream& frag_stream ){
	int fragmentI;
	fragment_list.clear();
	fragment_set.clear();
	while( frag_stream >> fragmentI ){
		fragment_list.push_back( fragmentI );
		fragment_set.insert( fragmentI );
	}
}

#ifdef WIN32

/**
 * Open a fragment list file (r/w) and read its contents,
 * return the file handle.  Leaves the file pointer at the end
 * of the file.  Returns INVALID_HANDLE_VALUE on error.
 */
HANDLE FragmentListFile::openAndReadWin32(){
	HANDLE fragfile;
	// this is really lame:  lock the file when opening
	// and if it's already open by another process, loop until
	// it can be opened
	while(true){
		fragfile = CreateFile( frag_filename.c_str(), 
			GENERIC_READ | GENERIC_WRITE,
			0, 0, OPEN_ALWAYS, 0, 0 );
		if( fragfile == INVALID_HANDLE_VALUE &&
			GetLastError() == ERROR_SHARING_VIOLATION )
						// the file is locked by another process
			Sleep(25);	// sleep 25 milliseconds and try again
		else
			break;	// we either opened it successfully or failed fatally
	}

	if( fragfile == INVALID_HANDLE_VALUE ){
		cerr << "Error opening " << frag_filename << endl;
		return INVALID_HANDLE_VALUE;	// failed
	}
	
	// determine the file size
	// read the whole file into a stringstream
	// parse file using c++ i/o functionality
	DWORD fragfile_size = GetFileSize( fragfile, 0 );
	char* buf = new char[ fragfile_size + 1 ];
	DWORD bytes_read = 0;
	if( !ReadFile( fragfile, buf, fragfile_size, &bytes_read, 0 ) ){
		// an error occurred... do something
		cerr << "Error reading " << frag_filename << endl;
		CloseHandle( fragfile );
		return INVALID_HANDLE_VALUE;
	}
	buf[ fragfile_size ] = 0;

	stringstream ff_sstream( buf );
	delete[] buf;
	parseStream( ff_sstream );
	return fragfile;
}

/**
 * Open a fragment list file and add a fragment if it doesn't
 * already exist in the file
 */
void FragmentListFile::addFragment( int fragmentI ){
	HANDLE fragfile = openAndReadWin32();

	if( fragfile == INVALID_HANDLE_VALUE ){
		// errored out
		return;
	}

	// add the fragment if its not already in the set
	if( fragment_set.find( fragmentI ) == fragment_set.end() ){
		// add it
		char fragno[8]; //space for the frag number
		int nchars = sprintf(fragno, "%03d\n", fragmentI);
		uint64 written;
		WriteFile( fragfile, fragno, nchars, &written, 0 );
	}

	CloseHandle( fragfile );
	fragment_set.insert( fragmentI );
	fragment_list.push_back( fragmentI );
}

#else

/**
 * Open a fragment list file (r/w) and read its contents,
 * return the file descriptor.  Leaves the file pointer at the end
 * of the file.  Returns -1 on error.
 */
int FragmentListFile::openAndReadPOSIX(){
	int mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH;
	int fragfile = open( frag_filename.c_str(), O_RDWR | O_CREAT, mode );
	if( fragfile < 0 ){
		cerr << "Error opening " << frag_filename << endl;
		perror( "open" );
		return -1;
	}
	if (debug_msg)
		LOG_MSG << "Locking fragment list" << endl;
	int lockval = getlock( fragfile, F_SETLKW, F_WRLCK );
	if ( lockval == -1 ){
		// error locking the file
		cerr << "Error locking file " << frag_filename << endl;
		perror( "getlock" );
		return -1;
	}
	if (debug_msg)
		LOG_MSG << "Locked fragment list" << endl;

	int fragfile_size = statFileSize( frag_filename.c_str() );
	
	char* buf = new char[ fragfile_size + 1 ];
	int bytes_read = read( fragfile, buf, fragfile_size );
	buf[ fragfile_size ] = 0;
	if( bytes_read != fragfile_size ){
		// an error occurred... do something
		cerr << "Error reading from " << frag_filename << endl;
		close( fragfile );
		return -1;
	}

	stringstream ff_sstream( buf );
	delete[] buf;
	parseStream( ff_sstream );
	return fragfile;
}

void FragmentListFile::addFragment( int fragmentI ){
	int fragfile = openAndReadPOSIX();

	if( fragfile == -1 ){
		// errored out
		return;
	}

	// add the fragment if its not already in the set
	if( fragment_set.find( fragmentI ) == fragment_set.end() ){
		// add it
		char fragno[5]; //space for the frag number
		int nchars = sprintf( fragno, "%03d\n", fragmentI );
		write( fragfile, fragno, nchars );
	}

	close( fragfile );
	fragment_set.insert( fragmentI );
	fragment_list.push_back( fragmentI );
}

#endif

bool FragmentListFile::contains( int fragmentI ){
	return fragment_set.find( fragmentI ) != fragment_set.end();
}

void FragmentListFile::setTimestamps( char* all_dates, int total_datelen ){
	char* tmp_date = all_dates;
	while( tmp_date != all_dates + total_datelen ){
		timestamps.push_back( tmp_date );
		tmp_date += strlen( tmp_date ) + 1;
	}
}

void FragmentListFile::checkTimestamps(){
	// if the remote frag has a different modification time than the local one
	// we pretend the local one doesn't exist
	vector< int > up_to_date_fragments;
	char frag_num_buff[4];
	for(int fragI = 0 ; fragI < (int)fragment_list.size(); fragI++){
		if( fragI >= timestamps.size() )
			break;	// the new database has fewer frags than the old...

		// always use 3 digit fragment ids
		sprintf(frag_num_buff,"%03d",fragment_list[ fragI ]);
		string local_fragname = config.localPath() + database_name + "." + frag_num_buff;
		char* m_dbname = (char*)malloc( local_fragname.length() + 1 );
		strcpy( m_dbname, local_fragname.c_str() );
		ReadDBFILEPtr rdfp = readdb_new_ex2( m_dbname, db_type == "p", READDB_NEW_DO_REPORT, NULL, NULL );
		free( m_dbname );
		if( rdfp == NULL ){
			cerr << "readdb unable to open " << local_fragname << endl;
			continue;
		}
		char* date_str = readdb_get_date( rdfp );

		if( date_str == timestamps[ fragment_list[ fragI ] ] ){
			up_to_date_fragments.push_back( fragment_list[ fragI ] );
		}else if( debug_msg ){
			LOG_MSG << "Out of date fragment:\n";
			LOG_MSG << "Local frag " << local_fragname << " created on " << date_str << endl ;
			LOG_MSG << "Remote frag " << frag_num_buff << " created on " << timestamps[ fragment_list[ fragI ] ] << endl ;
		}	

		rdfp = readdb_destruct( rdfp );
	}
	fragment_list = up_to_date_fragments;
	fragment_set = set< int >( fragment_list.begin(), fragment_list.end() );
}

void FragmentListFile::SendList( int dest )
{

	SendIntVec(fragment_list, dest, DB_FRAGMENTS_TAG);
}
	
vector<string> makeFragmentExtensions( string db_type);
const vector<string>& FragmentExtensions( const string& db_type)
{
  static vector<string> n_extensions = makeFragmentExtensions( "n" );
  static vector<string> p_extensions = makeFragmentExtensions( "p" );
  if( db_type == "n" ){
  	return n_extensions;
  }
  return p_extensions;
}

/**
 * Because DB "indices" are required, all 7 extensions have become
 * mandatory.
 */
vector<string> makeFragmentExtensions(string db_type)
{
	vector< string > extensions;
	
	extensions.push_back( "." + db_type + "hr" );
	extensions.push_back( "." + db_type + "in" );
	extensions.push_back( "." + db_type + "sq" );
	extensions.push_back( "." + db_type + "nd" );
	extensions.push_back( "." + db_type + "ni" );
	extensions.push_back( "." + db_type + "sd" );
	extensions.push_back( "." + db_type + "si" );
	
	return extensions;
}

