/******************************************************************************
** fragment_list.hpp
** Provides an API to the fragment list on worker nodes
** $Id: fragment_list.hpp,v 1.10 2005/06/20 03:58:52 koadman Exp $
** $Revision: 1.10 $
** 
** This file is a part of mpiBLAST and is copyright (c)
** by Aaron Darling of Los Alamos National Laboratory for the Regents
** of the University of California.
** 
** License:
** 
** Copyright 2002-2005. The Regents of the University of California.
** 
** This material was produced under U.S. Government contract
** W-7405-ENG-36 for Los Alamos National Laboratory, which is operated
** by the University of California for the U.S. Department of Energy.
** The Government is granted for itself and other acting on its behalf
** a paid-up, non-exclusive, irrevocable worldwide license in the
** material to reproduce, prepare derivative works, and perform
** publicly and display publicly. Beginning five (5) years after
** November 6, 2002, subject to additional five-year worldwide
** renewals, the Government is granted for itself and others acting on
** its behalf a paid-up, nonexclusive, irrevocable worldwide license
** in this material to reproduce, prepare derivative works, distribute
** copies to the public, perform publicly and display publicly, and to
** permit others to do so. NEITHER THE UNITED STATES NOR THE UNITED
** STATES DEPARTMENT OF ENERGY, NOR THE UNIVERSITY OF CALIFORNIA, NOR
** ANY OF THEIR EMPLOYEES, MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR
** ASSUMES ANY LEGAL LIABLITY OR RESPONSIBILITY FOR THE ACCURACY,
** COMPLETENESS, OR USEFULNESS OF ANY INFORMATION, APPARATUS, PRODUCT,
** OR PROCESS DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE
** PRIVATELY OWNED RIGHTS.
** 
** Additionally, this program is free software; you can distribute it
** and/or modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2 of
** the License, or any later version. Accordingly, this program is
** distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or
** FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
** License for more details. You should have received a copy of the
** GNU General Public License along with this program; if not, write
** to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
** Boston, MA 02111-1307, USA.
** 
** ------------------------------------------------------------
** 
** See the file 'COPYING' for the details of the GNU General Public License
** 
******************************************************************************/

#ifndef __MPI_BLAST_FRAGMENT_LIST
#define __MPI_BLAST_FRAGMENT_LIST

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string>
#include <vector>
#include <set>
#include <iostream>
#include "mpiblast_config.hpp"

/**
 * API for reading and modifying a database fragment list file.
 * The fragment list file tracks the database fragments currently on
 * local storage.
 */
class FragmentListFile {
public:
	FragmentListFile(){};
	/**
	 * Opens and reads a fragment list file, throws a (const char *) exception
	 * if something fails.
	 */
  	FragmentListFile( const std::string& filename, const MpiBlastConfig& config, const std::string& database_name, const std::string& db_type );
 	FragmentListFile( const FragmentListFile& dbsf )
	{
		*this = dbsf;
	};
	
  	FragmentListFile& operator=( const FragmentListFile& mbc );
  	~FragmentListFile()
	{
		// Do nothing
	};
  
	/**
	* Populate the fragment list from an input stream containing a fragment list
	*/
	void parseStream( std::istream& frag_stream );

#ifdef WIN32
	/**
	* Open a fragment list file (r/w) and read its contents,
	* return the file handle.  Leaves the file pointer at the end
	* of the file.  Returns INVALID_HANDLE_VALUE on error.
	*/
	void* openAndReadWin32();
#else
	/**
	* Open a fragment list file (r/w) and read its contents,
	* return the open file descriptor.  Leaves the file pointer at the end
	* of the file.  Returns -1 on error.
	*/
	int openAndReadPOSIX();
#endif

	/** 
  	 * Adds the fragment ID given in <code>fragmentI</code>
  	 * to the fragment list file
  	 */
  	void addFragment( int fragmentI );
	
	/**
	 * Set the database timestamps for the master copy of the database
	 * @param	all_dates	a series of null-terminated datestamp strings, 
	 *						one for every fragment in the FULL database
	 * @param	total_datelen	The total length of space allocated for all_dates
	 */
	void setTimestamps( char* all_dates, int total_datelen );

	/**
	 * Remove any out-of-date fragments from fragment_list and fragment_set
	 */
	void checkTimestamps();

	/**
	 * Send the fragment list to another node.  Uses sendIntVec()
	 */
	void SendList( int dest );
	
	/**
	 * Returns the number of fragments currently on local storage
	 */
	inline int fragmentCount()
	{
		return (int)(fragment_list.size());
	};
	
	/**
	 * Returns true if the fragment list contains fragmentI
	 */
	bool contains( int fragmentI );
private:
  	std::vector< int > fragment_list;
 	std::set< int > fragment_set;
  	std::string frag_filename;
	
	// these variables are for db updates,
	// ideally they should all be tucked inside
	// the MpiBlastConfig and there should be a
	// single per-process instance of MpiBlastConfig
	MpiBlastConfig config;
	std::string database_name;
	std::string db_type;
	std::vector< std::string > timestamps;
};


/**
 * File name extensions for sequence database files
 */
const std::vector< std::string >& FragmentExtensions( const std::string& db_type);

extern bool lock_mbf;	/**< When true use file-locking whenever accessing a fragment list */

#endif
