/******************************************************************************
** mpiblast.cpp
** The core of mpiBLAST
** $Id: mpiblast.cpp,v 1.85 2005/12/20 02:53:08 koadman Exp $
** $Revision: 1.85 $
** 
** This file is a part of mpiBLAST and is copyright (c)
** by Aaron Darling of Los Alamos National Laboratory for the Regents
** of the University of California.
** 
** License:
** 
** Copyright 2002-2005. The Regents of the University of California.
** 
** This material was produced under U.S. Government contract
** W-7405-ENG-36 for Los Alamos National Laboratory, which is operated
** by the University of California for the U.S. Department of Energy.
** The Government is granted for itself and other acting on its behalf
** a paid-up, non-exclusive, irrevocable worldwide license in the
** material to reproduce, prepare derivative works, and perform
** publicly and display publicly. Beginning five (5) years after
** November 6, 2002, subject to additional five-year worldwide
** renewals, the Government is granted for itself and others acting on
** its behalf a paid-up, nonexclusive, irrevocable worldwide license
** in this material to reproduce, prepare derivative works, distribute
** copies to the public, perform publicly and display publicly, and to
** permit others to do so. NEITHER THE UNITED STATES NOR THE UNITED
** STATES DEPARTMENT OF ENERGY, NOR THE UNIVERSITY OF CALIFORNIA, NOR
** ANY OF THEIR EMPLOYEES, MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR
** ASSUMES ANY LEGAL LIABLITY OR RESPONSIBILITY FOR THE ACCURACY,
** COMPLETENESS, OR USEFULNESS OF ANY INFORMATION, APPARATUS, PRODUCT,
** OR PROCESS DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE
** PRIVATELY OWNED RIGHTS.
** 
** Additionally, this program is free software; you can distribute it
** and/or modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2 of
** the License, or any later version. Accordingly, this program is
** distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or
** FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
** License for more details. You should have received a copy of the
** GNU General Public License along with this program; if not, write
** to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
** Boston, MA 02111-1307, USA.
** 
** ------------------------------------------------------------
** 
** See the file 'COPYING' for the details of the GNU General Public License
** 
******************************************************************************/

/**
 * 11/2004 A. E. Darling
 * - Restructured mpiBLAST to support query and database segmentation.
 *   The program now consists of three types of processes:  worker, 
 *   a scheduler, and a result writer.  The workers perform BLAST searches,
 *   and are assigned a query and a database fragment to search by the
 *   scheduler process.  When the worker completes a search it sends the 
 *   results asynchronously to the writer process and immediately requests
 *   another assignment from the scheduler (workers never block).  The program 
 *   completes when every combination of query and database fragment has been
 *   searched.
 * ROADMAP:
 *   Execution begins in main() with MPI initialization.  main() calls 
 *   MpiBlast::main() in which each process parses command line options.
 *   Rank 0 tree broadcasts the query file to all other processes.  
 *   Next, rank 0 uses the NCBI library to calculate the effective length
 *   of each query and the database (the effective search space).  Rank 0
 *   broadcasts these effective lengths so that correct E-value statistics can
 *   be calculated by workers for each search result.  Finally, rank 0
 *   broadcasts the timestamps for each database fragment to the other processes
 *   this allows other processes to check whether their copy of the database
 *   is up-to-date.
 *   Each process then assumes a role: writer, scheduler, or worker.  Typically
 *   ranks 0 and 1 will be the writer and scheduler, respectively.  All
 *   other processes assume worker roles.
 */
 
#include "mpiblast.hpp"

extern "C" {
#include "getopt.h"
#include "mpi_util.h"
#include "blast_hooks.h"
#include "readdb.h"
#include "distributed_bioseq.h"
}

using namespace std;

#ifdef MPE
#include <mpe.h>
#endif
int cpstart, cpend, blaststart, blastend, mergestart, mergeend; //MPE ids
int outputstart, outputend, recvbioseqstart, recvbioseqend; //more MPE ids

// These MPE ids were added to replace the profile_msg functionality
int init_blast_start, init_blast_end, load_queries_start, load_queries_end;
int cleanup_blast_start, cleanup_blast_end, send_results_start, send_results_end;
int send_frag_list_start, send_frag_list_end, file_cleanup_start, file_cleanup_end;
int bcast_filter_start, bcast_filter_end;
int bcast_query_start, bcast_query_end, bcast_qadj_start, bcast_qadj_end;
int receive_results_start, receive_results_end, worker_file_setup_start, worker_file_setup_end;
int init_ncbi_start, init_ncbi_end;

#ifndef WIN32
#include <signal.h>
#include <sys/resource.h>
#include <unistd.h>
#endif

// File local variables functions
string localPath;	/**< Used by cleanupDB() to give the local storage path */
static MpiBlast* mpiblast_ptr = NULL;
static bool running_blast = false;	/**< set to true when executing code in the NCBI library */
static bool sameSharedLocal = false;	/**< set to true when shared and local storage are identical */
void exitCleanly();
void terminateProgram( int sig );

/** The writer periodically syncs up with writers to avoid
    getting swamped */
int query_sync_interval = 5;

// DEBUG --  Electric fence global variables
//extern int EF_ALLOW_MALLOC_0;
//extern int EF_PROTECT_BELOW;



int main( int argc, char* argv[] )
{
	// DEBUG --  Electric fence global variables
	//EF_ALLOW_MALLOC_0 = 1;
	//EF_PROTECT_BELOW = 1;
	
	// Perform a quick version check (and exit the program if
	// the version flag is found). If we don't check now, there's no way for
	// the user to query the version with out having mpi up and running.
	for(int i = 1;i < argc;i++){
		if(strcmp(argv[i], "--version") == 0){
			cerr << argv[0] << " version " 
			     << VERSION << endl;
			exit(1);
		}
	}
	if (argc < 6) {
		cerr << "mpiBLAST requires the following options: -d [database] -i [query file] -p [blast program name]\n";
		exit(1);
	}

#ifdef MPE
	char mpelogfmt[20] = "MPE_LOG_FORMAT=SLOG";
	putenv(mpelogfmt);
#endif
	double realstarttime = clock();

	/* Start up MPI */
	
	// Original version (no MPI thread support)
	MPI_Init(&argc, &argv);
	
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &node_count);

	//Check for -p, -i, -d -- the required args
	int ArgsOK = 0;
	for(int i = 1;i < argc;i++){
		if(strcmp(argv[i], "-d") == 0)
			ArgsOK++;
		if(strcmp(argv[i], "-i") == 0)
			ArgsOK++;
		if(strcmp(argv[i], "-p") == 0)
			ArgsOK++;
		if ( (ArgsOK != 3) && (i == argc-1)){
			cerr << "mpiBLAST requires the following options: -d [database] -i [query file] -p [blast program name]\n";
			terminateProgram(-1);
		}
	}
	// set an error handler that won't abort
	MPI_Errhandler_set( MPI_COMM_WORLD, MPI_ERRORS_RETURN );
	
	//Make sure scheduler_process isn't > node_count
	//If it is, set it to 1. I see no reason to die on this error.
	if (scheduler_process > (node_count - 1)) 
		scheduler_process = 1;


	prog_start = MPI_Wtime();
	double progstarttime = clock();
#ifdef MPE
	cpstart = MPE_Log_get_event_number(); 
	cpend = MPE_Log_get_event_number(); 
	blaststart = MPE_Log_get_event_number(); 
	blastend = MPE_Log_get_event_number(); 
	mergestart = MPE_Log_get_event_number(); 
	mergeend = MPE_Log_get_event_number();
	outputstart = MPE_Log_get_event_number();
	outputend = MPE_Log_get_event_number();
	recvbioseqstart = MPE_Log_get_event_number();
	recvbioseqend = MPE_Log_get_event_number();
	if (rank==0){
		MPE_Describe_state(cpstart, cpend, "CopyFragment","DarkSlateGray");
		MPE_Describe_state(blaststart, blastend, "RunBlastall","DarkKhaki");
		MPE_Describe_state(mergestart, mergeend, "MergeResults","grey");
		MPE_Describe_state(outputstart, outputend, "outputResults","LightGreen");
		MPE_Describe_state(recvbioseqstart, recvbioseqend, "ReceiveBioseqs","pink");
	}
		
	// the following MPE ids replace profile log messages:
	init_blast_start = MPE_Log_get_event_number();
	init_blast_end = MPE_Log_get_event_number();
	load_queries_start = MPE_Log_get_event_number();
	load_queries_end = MPE_Log_get_event_number();
	cleanup_blast_start = MPE_Log_get_event_number();
	cleanup_blast_end = MPE_Log_get_event_number();
	send_results_start = MPE_Log_get_event_number();
	send_results_end = MPE_Log_get_event_number();
	send_frag_list_start = MPE_Log_get_event_number();
	send_frag_list_end = MPE_Log_get_event_number();
	file_cleanup_start = MPE_Log_get_event_number();
	file_cleanup_end = MPE_Log_get_event_number();
	bcast_query_start = MPE_Log_get_event_number();
	bcast_query_end = MPE_Log_get_event_number();
	bcast_filter_start = MPE_Log_get_event_number();
	bcast_filter_end = MPE_Log_get_event_number();
	bcast_qadj_start = MPE_Log_get_event_number();
	bcast_qadj_end = MPE_Log_get_event_number();
	receive_results_start = MPE_Log_get_event_number();
	receive_results_end = MPE_Log_get_event_number();
	worker_file_setup_start = MPE_Log_get_event_number();
	worker_file_setup_end = MPE_Log_get_event_number();
	init_ncbi_start = MPE_Log_get_event_number();
	init_ncbi_end = MPE_Log_get_event_number();
	
	if (rank ==0){
		MPE_Describe_state(init_blast_start, init_blast_end, "InitBLAST()","azure");
		MPE_Describe_state(load_queries_start, load_queries_end, "loadQueries()","MistyRose");
		MPE_Describe_state(cleanup_blast_start, cleanup_blast_end, "cleanupBLAST()","cornflower blue");
		MPE_Describe_state(send_results_start, send_results_end, "sendResults()","CadetBlue");
		MPE_Describe_state(send_frag_list_start, send_frag_list_end, "sendFragmentList","HotPink");
		MPE_Describe_state(file_cleanup_start, file_cleanup_end, "sendFragmentList","tomato");
		MPE_Describe_state(bcast_query_start, bcast_query_end, "bcast query file","grey30");
		MPE_Describe_state(bcast_qadj_start, bcast_qadj_end, "bcast query adjustments","MediumPurple4");
		MPE_Describe_state(receive_results_start, receive_results_end, "receiveResults()","sienna1");
		MPE_Describe_state(worker_file_setup_start, worker_file_setup_end, "worker file setup","plum1");
		MPE_Describe_state(init_ncbi_start, init_ncbi_end, "initNCBI()","gray90");
	}
#endif

	log_stream = &cerr;
	MpiBlast mpb;
	
	if(node_count < 3){ 
		//sorry, we need at least 3 processes.
		terminateProgram(0);
	}
#ifndef WIN32
	signal( SIGINT, terminateProgram );
	signal( SIGTERM, terminateProgram );
	signal( SIGSEGV, terminateProgram );
#endif

	// register an exit function to abort MPI if blast exits
	// without an error code
	atexit( &exitCleanly );
	
	mpiblast_ptr = &mpb;
	
	int retcode = mpb.main( argc, argv );
	
	if( debug_msg ){
		LOG_MSG << "rank " << rank << " exiting successfully\n";
		LOG_MSG << "MPI startup  time is " << (double)((progstarttime - realstarttime)/CLOCKS_PER_SEC) << endl;
	}
	
	mpiblast_ptr = NULL;
	MPI_Finalize();
	return retcode;
}


void deleteRegisteredFiles();
void deleteRegisteredFiles() {
	// don't be a slob, clean up after yourself:
	// delete any files that are laying around
	vector< string >& del_files = registerFileToDelete();
	for( int fileI = 0; fileI < del_files.size(); fileI++ )
		removeFile( del_files[ fileI ] );
	del_files.clear();	// clear the deleted files from the list
}

/** 
 * This function is registered with atexit() to run before mpiBLAST terminates.
 * It deletes files that need to be deleted and calls terminateProgram() if
 * the NCBI library tried to abort execution
 */
void exitCleanly(){
	deleteRegisteredFiles();
	// abort MPI if blast crashed
	if( running_blast )
		terminateProgram( -1 );
}

/**
 * Aborts the mpiBLAST program using MPI_Abort()  On unix this function gets 
 * registered to handle interrupt, terminate, and segfault signals.
 * @param sig	A unix signal to die with.  Special case is 0, indicating that mpiblast is being run with too few processes.
 */
void terminateProgram( int sig ){
	if(sig==0) 
		cerr << "Sorry, mpiBLAST must be run on 3 or more nodes\n";
	else {
		cerr << rank << "\t" << (MPI_Wtime()-prog_start) 
		     << "\tBailing out with signal "<< sig << endl;
		if( mpiblast_ptr != NULL )
			if( mpiblast_ptr->remove_db ){
				deleteMpiBlastFiles( localPath );
				removeDirectory( localPath );
			}
	}
	MPI_Abort( MPI_COMM_WORLD, 0 );
}

#ifdef WIN32
BOOL Win32CtrlHandler( DWORD fdwCtrlType ) 
{ 
	cerr << "Caught a signal: " << fdwCtrlType << endl;
  switch( fdwCtrlType ) 
  { 
    // Handle the CTRL-C signal. 
    case CTRL_C_EVENT: 
      cerr << "Ctrl-C event\n\n";
      return( TRUE );
 
    // CTRL-CLOSE: confirm that the user wants to exit. 
    case CTRL_CLOSE_EVENT: 
      cerr << "Ctrl-Close event\n\n";
      return( TRUE ); 
 
    case CTRL_LOGOFF_EVENT: 
      cerr << "Ctrl-Logoff event\n\n";
      return TRUE; 

    case CTRL_BREAK_EVENT: 
      cerr << "Ctrl-Break event\n\n";
      return TRUE; 
 
    case CTRL_SHUTDOWN_EVENT: 
      cerr << "Ctrl-Shutdown event\n\n";
      return TRUE; 
 
	// Pass other signals to the next handler. 
    default: 
      return FALSE; 
  } 
} 
#endif

MpiBlast::~MpiBlast(){ 
	log_stream = &cerr;
}


/**
 *  Scans through the list of SeqAlignPtrs separating it into several lists,
 *  each of which corresponds to a single BLAST database hit.
 */
void BreakUpResults( vector< pair< int, SeqAlignPtr > >& results_vec, SeqAlignPtr sap, int searched_frag_id ){
	SeqAlignPtr cur_a = sap;
	results_vec.clear();
	results_vec.push_back( pair< int, SeqAlignPtr >( searched_frag_id, cur_a ) );
	int cur_count = 1;
	//	cerr << cur_count << "\t" << "cur_a: " << cur_a << endl;
	while( cur_a->next != NULL ){
		//		cerr << cur_count << "\t" << "cur_a: " << cur_a->next << endl;
		SeqIdPtr cur_id = TxGetSubjectIdFromSeqAlign( cur_a ); 
		SeqIdPtr next_id = TxGetSubjectIdFromSeqAlign( cur_a->next );
		if( SeqIdComp( cur_id, next_id ) != SIC_YES ){
			// break the chain at cur_a, add cur_a->next
			SeqAlignPtr tmp_next = cur_a->next;
			cur_a->next = NULL;
			//			cerr << "tmp_next\t" << tmp_next << endl;
			results_vec.push_back( pair< int, SeqAlignPtr >( searched_frag_id, tmp_next ) );
			cur_a = tmp_next;
			//			cerr << "Segment count is: " << cur_count << endl;
			cur_count = 1;
		}else{
			cur_a = cur_a->next;
			cur_count++;
		}
	}
}


/**
 * Mergesorts new results with existing results for a single query
 */
void mergeSeqAlignResults( const vector< pair< int, SeqAlignPtr > >& results, const vector< pair< int, SeqAlignPtr > >& new_results, 
			  vector< pair< int, SeqAlignPtr > >& merged_results ){
	uint resultI = 0;
	uint resultJ = 0;
	while( resultI != results.size() && resultJ != new_results.size() ){

		int a_score = 0, b_score = 0, a_number, b_number;
		Nlm_FloatHi a_bit_score = 0, b_bit_score = 0, a_evalue, b_evalue;

		SeqAlignPtr sapI = results[ resultI ].second;
		SeqAlignPtr sapJ = new_results[ resultJ ].second;

		GetScoreAndEvalue( sapI, &a_score, &a_bit_score, &a_evalue, &a_number );
		GetScoreAndEvalue( sapJ, &b_score, &b_bit_score, &b_evalue, &b_number );
//		cerr << "a_score: " << a_bit_score << "\t b_score: " << b_bit_score << endl;
		if( a_evalue < b_evalue ||
			a_evalue == b_evalue && a_bit_score > b_bit_score ||
			a_evalue == b_evalue && a_bit_score == b_bit_score && results[ resultI ].first > new_results[ resultJ ].first)
		{
			merged_results.push_back( results[ resultI ] );
			resultI++;
		}else{
			merged_results.push_back( new_results[ resultJ ] );
			resultJ++;
		}
	}
	for( ; resultI != results.size(); resultI++ ){
		merged_results.push_back( results[ resultI ] );
	}
	for( ; resultJ != new_results.size(); resultJ++ ){
		merged_results.push_back( new_results[ resultJ ] );
	}
}


void mergeResults( map< int, vector< pair< int, SeqAlignPtr > > >& query_map, vector< SeqAlignPtr >& new_results, int searched_frag_id, int* result_list )
{
	for( uint resultI = 0; resultI < new_results.size(); resultI++ ){

		int query_id = result_list[ resultI ];
		vector< pair< int, SeqAlignPtr > > cur_results;
		BreakUpResults( cur_results, new_results[ resultI ], searched_frag_id );

		map< int, vector< pair< int, SeqAlignPtr > > >::iterator query_iter = query_map.find( query_id );
		
		if( query_iter == query_map.end() ){
			// no results for this query yet.  add it.
			query_map.insert( map< int, vector< pair< int, SeqAlignPtr > > >::value_type( query_id, cur_results ) );
		}else{
			// merge these results with the existing query results.
			vector< pair< int, SeqAlignPtr > > merged_results;
			mergeSeqAlignResults( query_iter->second, cur_results, merged_results );
			query_map.erase( query_iter );
			query_map.insert( map< int, vector< pair< int, SeqAlignPtr > > >::value_type( query_id, merged_results ) );
		}
	}
}


/**
  * Asynchronously notify the writer process that a search has completed.
  * @param dest        The rank of the process to notify
  * @param retcode     The return code of the blast executable
  * @param queries     The query IDs that were searched
  */
void MpiBlast::notifyWriterOfCompletion( int dest, int frag_id, vector< int >& queries ){

	// try freeing any existing send buffers before sending more data
	freeCompletedSendBuffers();
	
	// construct a message of the form:
	// worker->writer:  result_id, frag_id, searched count, searched list, success count, success list, data size
	// send list of queries that had hits
	int success_count = ValNodeLen( result_id_list );
	int message_size = 3 + queries.size() + 1 + success_count + 1;
	int* message_buf = NULL;
	meta_MPI_Alloc_mem(sizeof( int )*message_size, MPI_INFO_NULL, &message_buf);

	// pack the message buf
	int bufI = 0;
	message_buf[ bufI++ ] = result_id;
	message_buf[ bufI++ ] = frag_id;
	message_buf[ bufI++ ] = queries.size();
	for( uint queryI = 0; queryI < queries.size(); queryI++ )
		message_buf[ bufI++ ] = queries[ queryI ];

	message_buf[ bufI++ ] = success_count;
	ValNodePtr cur_result_id = result_id_list;
	for( uint queryI = 0; queryI < success_count; queryI++ ){
		message_buf[ bufI++ ] = cur_result_id->data.intvalue;
		cur_result_id = cur_result_id->next;
	}
	
	// add result size
	int result_size = 0;
	ValNodePtr cur_result_bsp = result_bsp_list;
	while( cur_result_bsp != NULL ){
		result_size += BSLen((ByteStorePtr)(cur_result_bsp->data.ptrvalue));
		cur_result_bsp = cur_result_bsp->next;
	}
	if( debug_msg ){
		LOG_MSG << "message_buf[" << bufI << "] will indicate " << result_size << " bytes of results\n";
	}
	message_buf[ bufI++ ] = result_size;
	
	// finally, send the completion message along to the writer
// async send
	MPI_Request* req = (MPI_Request*)malloc( sizeof( MPI_Request ) );
	memset( req, 0, sizeof( MPI_Request ) );
	MPI_Issend( message_buf, message_size, MPI_INT, dest, BLAST_RETCODE_TAG, MPI_COMM_WORLD, req );
	addSendBuffer( message_buf, req );

// sync send
//	MPI_Ssend( message_buf, message_size, MPI_INT, dest, BLAST_RETCODE_TAG, MPI_COMM_WORLD );

	// prepare the result buffer to go over "the wire"
	// and put it in the result ID map
	// the writer will later request this buffer by its ID
	// allocate space
	unsigned char* buffer = NULL;
	meta_MPI_Alloc_mem(result_size, MPI_INFO_NULL, &buffer);

	// copy results into buffer space
	ValNodePtr cur_bsp = result_bsp_list;
	unsigned char* cur_buf = buffer;
	while( cur_bsp != NULL ){
		/* Copy the byte store to our mpi memory buffer */
		ByteStorePtr bs_ptr = (ByteStorePtr)cur_bsp->data.ptrvalue;
		BSMerge(bs_ptr, cur_buf);
		cur_buf += BSLen(bs_ptr);
		cur_bsp = cur_bsp->next;
		/* Free the byte store */
		BSFree(bs_ptr);
	}

	/* add the buffer to the result list */
	pair< int, unsigned char* > size_buf( result_size, buffer );
	result_id_data_map.insert( map< int, pair< int, unsigned char* > >::value_type( result_id, size_buf ) );
	result_id++;

	ValNodeFree( result_bsp_list );
	result_bsp_list = NULL;
	ValNodeFree( result_id_list );
	result_id_list = NULL;
}

// Send the results of the blast search 
// Note that if records are returned to
// the master one at a time, MPI becomes prone to the "errno=110" or "Connection timed
// out" error when running on linux (at least with kernel 2.4.18). The problem occurs
// with both mpich (1.2.5) and LAM MPI (6.5.9). This error is
// allegedly with the linux tcp/ip implementation. I have found that this problem is
// sensitive to the *NUMBER* of messages sent (in addition to the size of each
// message). The scheme of collecting results into "chunks" before sending to the
// master seems to correct the problem. More work may be required if the problem
// returns.

void MpiBlast::receiveResults( ResultData_t& rd, vector<SeqAlignPtr>& results )
{
	// the results buffer is structured as
	// Query n:
	//		SeqAlign1
	//		...
	//		SeqAlignN
	//		PartialBioseq1
	//		...
	//		PartialBioseqJ
	//		Stats
	// Query n+1:
	//		SeqAlign1
	//		...
	//		SeqAlignN
	//		PartialBioseq1
	//		...
	//		PartialBioseqJ
	//		Stats
	// etc...
	
	// Allocate space for the incoming data
	results.clear();
	
	int frag_id = rd.notification[1];
	int searched_count = rd.notification[2];
	int success_count = rd.notification[3+searched_count];
	int result_size = rd.notification[3+searched_count+1+success_count];
	if( debug_msg) {
		LOG_MSG << "receiveResults to read " << success_count << " records" << endl;
	}
	for( int qI = 0; qI < searched_count; qI++ ){
		int successI = 0;
		int queryI = rd.notification[3+qI];
		for( ; successI < success_count; successI++ ){		
			if( rd.notification[4+searched_count+successI] == queryI )
				break;
		}

		// Any results are in ASN.1 binary format
		AsnIoMemPtr aimp;
		if( use_binary_asn )
			aimp = AsnIoMemOpen( "rb", rd.buf, result_size );
		else
			aimp = AsnIoMemOpen( "r", rd.buf, result_size );

		if(aimp == NULL){
			throw __FILE__ "(MpiBlast::receiveResults): AsnIoMemOpen has failed!";
		}

		if( successI < success_count ){
			// this query had at least one hit
			SeqAnnotPtr sap = (SeqAnnotPtr)SeqAnnotAsnRead(aimp->aip, NULL);
			// Save the data
			results.push_back( (SeqAlignPtr)sap->data );		
			// Free the un-needed memory associated with the SeqAnnot structure (but protect the
			// SeqAlign structure contained within).
			sap->data = NULL;
			sap = SeqAnnotFree(sap);
										
			if( debug_msg) {
				LOG_MSG << "Read SeqAnnot\n";
			}

			// only receive bioseqs if the alignments will actually get used
			// e.g. only if output format < 10
	//		if( output_format < 10 && !disable_mpi_db ){
				if( debug_msg )
					LOG_MSG << "Reading bioseqs for frag/query " << frag_id << "/" << queryI << "\n";
				meta_MPE_Log_event(recvbioseqstart,0,"start recvBioseqs");
				readPartialBioseqs( results[ results.size() - 1 ], aimp, queryI );
				meta_MPE_Log_event(recvbioseqend,0,"end recvBioseqs");
	//		}
		}
		// read the stats
		AsnTypePtr atp;
		DataVal dv;
		for( int statI = 0; statI < 12; statI++ ){
			// read "Stat-data" struct opening
			atp = AsnReadId( aimp->aip, NULL, &statset_type );
			AsnReadVal( aimp->aip, atp, &dv );
			// read each stat-data value
//			at.type = &at;
			atp = AsnReadId( aimp->aip, NULL, &statdata_type );
			AsnReadVal( aimp->aip, atp, &dv );
			stats_array[queryI][statI] += dv.bigintvalue;
			// read "Stat-data" struct close
			atp = AsnReadId( aimp->aip, NULL, &statset_type );
			AsnReadVal( aimp->aip, atp, &dv );
		}

		// Clean up 
		AsnIoMemClose(aimp);
	}

}

void MpiBlast::purgeBioseqCache( int queryI ){
	ValNodePtr bs_ptr = query_data[queryI].bioseq_cache;
	while( bs_ptr != NULL ){
		// for some reason the NCBI library doesn't
		// seem to unlock Bioseqs after writing XML format
		// output.  Further, it seems that some bioseqs have multiple
		// locks while others have none.  Could do a hack here 
		// which looks up the object, checks the number of locks and
		// unlocks the appropriate number of times.
		// or could just fix and report the bug in the ncbi lib...
		// for now, punt on it and leak memory--the alternative is
		// to try to unlock anyways, which can result in crashes
//		if( output_format == 7 ){
			// call BioseqUnlock the appropriate number of times
//			BioseqUnlock( (BioseqPtr)bs_ptr->data.ptrvalue );
//		}
		if( output_format != 7 ){
			BioseqFree( (BioseqPtr)bs_ptr->data.ptrvalue );
		}
		bs_ptr = bs_ptr->next;
	}
	ValNodeFree( query_data[queryI].bioseq_cache );
	query_data[queryI].bioseq_cache = NULL;
}


void MpiBlast::readPartialBioseqs( SeqAlignPtr sap_p, AsnIoMemPtr aimp, int queryI ){
	while( sap_p != NULL ){

		// read subject bioseq
		BioseqPtr partial_seq;
		partial_seq = BioseqAsnRead(aimp->aip, NULL);

		// read the original bioseq length
		DataVal dv;
		// read open struct
		AsnTypePtr atp = AsnReadId( aimp->aip, NULL, &lenset_type );
		AsnReadVal( aimp->aip, atp, &dv );
		// read original length
		atp = AsnReadId( aimp->aip, NULL, &len_type );
		AsnReadVal( aimp->aip, atp, &dv );
		Int4 original_length = dv.bigintvalue;
		// read close struct
		atp = AsnReadId( aimp->aip, NULL, &lenset_type );
		AsnReadVal( aimp->aip, atp, &dv );

		// partial_seq is a temporary "clipboard" bioseq
		// don't let the SeqMgr index it...
		SeqMgrWriteLock();
		SeqMgrDeleteFromBioseqIndex( partial_seq );
		SeqMgrUnlock();
		
		// extract the coordinates of the aligned region
		Int4 start = -1;
		Int4 stop = -1;
		Uint1 strand = 0;
		SeqIdPtr subject_id = TxGetSubjectIdFromSeqAlign( sap_p );
		get_align_ends( sap_p, subject_id, &start, &stop, &strand );
		if( strand == 2 )
			strand = 1;	// the forward strand is always copied and sent

		// add to cache
		ValNodeAddPointer( &(query_data[ queryI ].bioseq_cache), 0, partial_seq );
		query_data[ queryI ].bioseq_lengths.push_back( original_length );
		query_data[ queryI ].bioseq_starts.push_back( start );
		query_data[ queryI ].bioseq_strands.push_back( strand );

		sap_p = sap_p->next;
	}
}

void MpiBlast::expandPartialBioseqs( int queryI ){
	ValNodePtr cur_cache = query_data[ queryI ].bioseq_cache;
	ValNodePtr new_cache = NULL;
	vector< int > new_lengths;
	int bioseqI = 0;
	while( cur_cache != NULL ){
		// receive bioseq
		BioseqPtr partial_seq = (BioseqPtr)cur_cache->data.ptrvalue;
		// receive the original bioseq length
		Int4 original_length = query_data[ queryI ].bioseq_lengths[ bioseqI ];
		
		// extract the coordinates of the aligned region
		Int4 start = query_data[ queryI ].bioseq_starts[ bioseqI ];
		Int4 stop = partial_seq->length + start - 1;
		Uint1 strand = query_data[ queryI ].bioseq_strands[ bioseqI ];

		// first check whether a partial bioseq for this SeqId already
		// exists in our mpiBLAST bioseq cache.  If not we can simply
		// add the new sequence to the cache.  If so, we need to write
		// in the part just received.

		BioseqPtr subject_bioseq = NULL;
		ValNodePtr cur_bs = new_cache;
		while( cur_bs != NULL ){
			if( SeqIdComp( ((BioseqPtr)cur_bs->data.ptrvalue)->id, partial_seq->id ) == SIC_YES ){
				/* this is the one we're looking for! */
				subject_bioseq = (BioseqPtr)cur_bs->data.ptrvalue;
			}
			cur_bs = cur_bs->next;
		}
		
		if( subject_bioseq == NULL ){
			// Insert junk before start.
			Int4 len = start;
			char* junk = (char*)malloc( len );
			BSSeek( partial_seq->seq_data, 0, SEEK_SET );
			BSInsert( partial_seq->seq_data, junk, len );
			partial_seq->length += len;
			free(junk);

			// add to cache
			ValNodeAddPointer( &new_cache, 0, partial_seq );
			new_lengths.push_back( original_length );
		}
		else
		{	
			// Add data to a pre-existing sequence in our cache
			// add any additional data up to the start...
			Int4 len = start - subject_bioseq->length;
			if( len > 0 ){
				char* junk = (char*)malloc( len );
				BSSeek( subject_bioseq->seq_data, subject_bioseq->length, SEEK_SET );
				BSInsert( subject_bioseq->seq_data, junk, len );
				subject_bioseq->length += len;
				free( junk );
			}

			// set the partial bioseq id to a temporary id so that the
			// seqmgr doesn't get confused when both have the same ID
			ValNodePtr tmp = ValNodeNew(NULL);
			tmp->choice = SEQID_LOCAL;
			ObjectIdPtr oid = ObjectIdNew();
			tmp->data.ptrvalue = (void*)oid;
			oid->str = StringSave("Clippy");
			SeqIdSetFree( partial_seq->id );
			partial_seq->id = tmp;

			// add the bioseqs to the SeqMgr index so that
			// editing functions can find them...
			SeqMgrWriteLock();
			SeqMgrAddToBioseqIndex( subject_bioseq );
			SeqMgrAddToBioseqIndex( partial_seq );
			SeqMgrUnlock();

			if( start < subject_bioseq->length ){
				// delete the portion of the bioseq that was just sent...
				Int4 del_end = min( stop, subject_bioseq->length - 1 );
				if( !BioseqDelete( subject_bioseq->id, start, del_end, FALSE, FALSE ) ){	
					LOG_MSG << "Error deleting bioseq!!\n";
				}
			}
			// Insert the sequence just sent
			if( !BioseqInsert( partial_seq->id, 0, partial_seq->length - 1, strand, subject_bioseq->id, start, FALSE, FALSE, FALSE ) ){
				LOG_MSG << "Error inserting bioseq!!\n";
			}
			
			// remove the bioseqs from the SeqMgr index...
			SeqMgrWriteLock();
			SeqMgrDeleteFromBioseqIndex( subject_bioseq );
			SeqMgrDeleteFromBioseqIndex( partial_seq );
			SeqMgrUnlock();

			// delete the partial sequence
			BioseqFree( partial_seq );
		}

		cur_cache = cur_cache->next;
		bioseqI++;
	}

	// now that all bioseqs are expanded, set their original lengths
	bioseqI = 0;
	cur_cache = new_cache;
	while( cur_cache != NULL ){
		((BioseqPtr)cur_cache->data.ptrvalue)->length = new_lengths[ bioseqI ];
		cur_cache = cur_cache->next;
		bioseqI++;
	}
	// free some memory
	query_data[ queryI ].bioseq_lengths.clear();
	query_data[ queryI ].bioseq_starts.clear();
	query_data[ queryI ].bioseq_strands.clear();
	ValNodeFree( query_data[ queryI ].bioseq_cache );

	// finally, set the cache to the new cache of expanded bioseqs
	query_data[ queryI ].bioseq_cache = new_cache;
}

/**
 * Write out the next query if it has been completely searched
 * The list of bioseqs referenced in the query are stored in a query-specific
 * bioseq cache.  We register a custom bioseq lookup function so that when
 * the NCBI lib calls BioseqLockById() during result output our custom function
 * can return the appropriate bioseq.
 */
void MpiBlast::outputQueryResults() {
	// loop while there are queries to output
	while( output_queryI < query_data.size() ){

	// check whether the next output query has been completely searched
	if( query_data[ output_queryI ].searched_frags != total_frags )
		return;	// this query isn't done yet
	
	if( debug_msg ){
		LOG_MSG << query_data[ output_queryI ].searched_frags << " / " << total_frags;
		CONT_LOG_MSG << " frags have been searched for query " << output_queryI << ".  Writing results\n";
	}
	
	expandPartialBioseqs( output_queryI );

	// print out the results
	if( !disable_mpi_db && !Enable_MPI_Distributed_DB()){
		cerr << "Error enabling mpiBLAST cached DB lookups" << endl;
		throw __FILE__ ":master: Error enabling mpiBLAST cached DB lookups";
	}
	
	MpiBlastEnableBioseqFetch();
	
	map< int, vector< pair< int, SeqAlignPtr > > >::iterator result_iter = results_map.find( output_queryI );
	SeqAlignPtr cur_result_head = NULL;

	if( result_iter == results_map.end() )
		outputResults( NULL, output_queryI ); // there were no results for this query
	else{
		// link up all the results for this query into a single list
		cur_result_head = (*result_iter).second[0].second;
		SeqAlignPtr cur_result_tail = cur_result_head;
		
		while(cur_result_tail->next != NULL){
			cur_result_tail = cur_result_tail->next;
		}
			
		for( uint resultI = 1; resultI < (*result_iter).second.size(); resultI++ ){
			cur_result_tail->next = (*result_iter).second[ resultI ].second;
			
			while(cur_result_tail->next != NULL){
				cur_result_tail = cur_result_tail->next;
			}
		}
		
		// output the list of results
		running_blast = true;
		if( !disable_mpi_db ){
			// if we're using the distributed db, then set the bioseq cache
			// for this query
			setBioseqCache( query_data[ output_queryI ].bioseq_cache );
		}

		if( debug_msg )
			LOG_MSG << "outputResults()\n";
		outputResults( cur_result_head, output_queryI );

		if( !disable_mpi_db ){
			setBioseqCache( NULL );
		}
		running_blast = false;

		// free memory
		results_map.erase( result_iter );
	}
	
	MpiBlastDisableBioseqFetch();
	
	if(!disable_mpi_db && !Disable_MPI_Distributed_DB()){
		LOG_MSG << "Error disabling MPI distributed DB" << endl;
	}
	
	if( debug_msg ) {
		LOG_MSG << "Finished output of query " << output_queryI << endl;
	}
	
	// free memory associated with this result!
	purgeBioseqCache( output_queryI );
	
	output_queryI++;
	
	} // end while loop over all queries
}

// JA -- send fragments over MPI
bool MpiBlast::sendFragsViaMPI(int fragment_id, int worker_id){

	const vector<string>& extensions = FragmentExtensions(db_type);

	char fragid[8];
	memset( fragid, 0, 8 );
	sprintf( fragid, "%03d", fragment_id );

	string copy_fragment_name;
	string copy_src;

	vector< string > copy_files;
	vector< string > ext;

	// determine which frags to copy and how many
	for( uint extI = 0; extI < extensions.size(); extI++ ){
		copy_fragment_name = database_name + "." + fragid + extensions[ extI ];
		copy_src = config.sharedPath() + copy_fragment_name;
		if( doesFileExist( copy_src ) ){
			ext.push_back( extensions[ extI ] );
			copy_files.push_back( copy_src );
		}
	}
	
	// tell the worker how many files it's getting
	int file_count = copy_files.size();
	MPI_Send( &file_count, 1, MPI_INT, worker_id, DB_DATATOSEND_TAG, MPI_COMM_WORLD );

	// send each filename extension and file
	for( uint fileI = 0; fileI < copy_files.size(); fileI++ ){
		char ext_name[5];
		strncpy( ext_name, ext[ fileI ].c_str(), 5 );
		MPI_Send( ext_name, ext[ fileI ].length() + 1, MPI_CHAR, worker_id, DB_EXTENSION_DATA_TAG, MPI_COMM_WORLD );
		sendFile( copy_files[ fileI ], worker_id, DB_DATATOSEND_TAG );
	}

	return true;
}


// JA -- receive fragments over MPI
bool MpiBlast::recvFragsViaMPI( int fragment_id ){

	char fragid[8];
	memset( fragid, 0, 8 );
	sprintf( fragid, "%03d", fragment_id );

	string copy_fragment_name;
	string copy_dest;

	if( debug_msg )	LOG_MSG << "Requesting fragment " << fragment_id << " from rank 0\n";
	// tell rank 0 which fragment we want to copy
	int f_tmp = fragment_id;
	MPI_Send( &f_tmp, 1, MPI_INT, 0, DB_DATATOSEND_TAG, MPI_COMM_WORLD );

	if( debug_msg )	LOG_MSG << "Getting frag file count for " << fragment_id << " from rank 0\n";
	// get file count from rank 0
	int file_count;
	MPI_Recv( &file_count, 1, MPI_INT, 0, DB_DATATOSEND_TAG, MPI_COMM_WORLD, MPI_STATUS_IGNORE );

	if( debug_msg )	LOG_MSG << "Got frag file count " << file_count << " from rank 0\n";
	
	for( uint fileI = 0; fileI < file_count; fileI++ ){
		char ext[5];
		MPI_Recv( ext, 5, MPI_CHAR, 0, DB_EXTENSION_DATA_TAG, MPI_COMM_WORLD, MPI_STATUS_IGNORE );
		if( debug_msg )	LOG_MSG << "Got frag extension name " << ext << " from rank 0\n";
		copy_fragment_name = database_name + "." + fragid;
		copy_fragment_name += ext;
		copy_dest = config.localPath() + copy_fragment_name;
		recvFile( copy_dest, 0, DB_DATATOSEND_TAG );
	}

	return true;
}

void MpiBlast::processResultReceiveCompletions()
{
	list<ResultData>::iterator r_iter = result_data.begin();
	while( r_iter != result_data.end() ){
		int flag = 0;
		MPI_Status status;
		MPI_Test( r_iter->req, &flag, &status );
		if(!flag){
			r_iter++;
			continue;
		}

		// give names to the completion message variables
		int* notification = r_iter->notification;
		int frag_id = notification[1];
		int searched_count = notification[2];
		int success_count = notification[3+searched_count];

		if( debug_msg )
			LOG_MSG << "Processing result_id: " << notification[0] << endl;

		// a fragment has been searched against a query, 
		// increment the appropriate searched_frags
		int bufI = 3;
		for( uint queryI = 0; queryI < searched_count; queryI++ ){
			query_data[ notification[bufI++] ].searched_frags++;
			if( debug_msg )
				LOG_MSG << "Query " << notification[bufI-1] << " now has " << query_data[ notification[bufI-1] ].searched_frags << " / " << total_frags << " fragments searched\n";
		}

		meta_MPE_Log_event(receive_results_start,0,"start receiving results");

		// receive the list of successful queries
		if( debug_msg )
			LOG_MSG << "Got result count of " << success_count << endl;
		if( success_count > 1 )
			LOG_MSG << "ARRGH!  more than one query is getting its results sent at a time!\n";

		vector< SeqAlignPtr > cur_results;
		running_blast = true;
		receiveResults( *r_iter, cur_results );
		running_blast = false;
			
		if( debug_msg ){
			LOG_MSG << "Receive was successful -- about to merge\n";
		}
		
		if( cur_results.size() > 0 ){
			// merge the results into a single list
			meta_MPE_Log_event(mergestart,0,"start merge");
			mergeResults( results_map, cur_results, frag_id, &(notification[4+searched_count]) );
			meta_MPE_Log_event(mergeend,0,"end merge");
		}
		if( debug_msg ){
			LOG_MSG << "Query results have been merged\n";
		}

		// erase this one since it has been processed
		free( r_iter->notification );
		meta_MPI_Free_mem( r_iter->buf );
		free( r_iter->req );
		list<ResultData>::iterator erase_iter = r_iter;
		r_iter++;
		result_data.erase( erase_iter );
	}
}

/**
 * The writer process waits for search results from search processes
 * Each time it receives results it tries to output them using the
 * outputResults() function
 */
void MpiBlast::writer()
{
	MPI_Status status;
	output_queryI = 0;
	
	// placeholder FragmentDistribution
	FragmentDistribution frag_dist( total_frags, node_count );
	// placeholder BlastJob
	BlastJob* bj = new BlastJob( total_frags, frag_dist );
	
	// QueryData vector to track whether queries have completed yet
	QueryData qd( bj, NULL );
	query_data = vector< QueryData >( query_count, qd );
	
	// don't write out the header if we're resuming
	if( !resume_run )
		outputHtmlHeader();

	// loop waiting for search results to come in
	// ...

	while( true ){

		//
		// wait for either a message from a worker or a result to be received
		//
		MPI_Status probe_status;
		int flag = 0;
		while( !flag ){
			MPI_Iprobe( MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &flag, &probe_status );
			
			if( flag ){
				// Check the message tag that we have probed to prevent a race condition 
				// (i.e. receiving BLAST_RESULTS_TAG before a BLAST_RETCODE_TAG) from 
				// causing problems.
				if( (probe_status.MPI_TAG == DB_DATATOSEND_TAG) || (probe_status.MPI_TAG == BLAST_RETCODE_TAG) ){
					// We only break for one of these two messages (i.e. if we get a BLAST_RESULTS_TAG, we
					// need to stay within this while loop).
					break;
				}
				
				// The message tag was not DB_DATATOSEND_TAG or BLAST_RETCODE_TAG, so reset the flag
				flag = 0;
			}

			// check whether results have been received that need to be 
			// processed and merged
			int prev_result_count = result_data.size();
			processResultReceiveCompletions();

			if( prev_result_count != result_data.size() )
			{
				// a result was received
				// check whether there are any complete query results to write out
				meta_MPE_Log_event(outputstart,0,"start output");
				outputQueryResults();
				meta_MPE_Log_event(outputend,0,"end output");
				if( output_queryI == query_count )
					break;
			}else{
				// nothing was received, sleep briefly to avoid a busy wait
#ifdef WIN32
				Sleep(10);
#else
				usleep(10000);
#endif
			}
			freeCompletedSendBuffers();
		}
		// if we're done writing then bail out!
		if( output_queryI == query_count )
			break;

		// check whether this is a worker trying to copy a fragment
		// via MPI.  This needs to be done by the rank 0 node
		// which ought to be the writer.
		if( probe_status.MPI_TAG == DB_DATATOSEND_TAG ){
			if( debug_msg ) LOG_MSG << "Worker " << probe_status.MPI_SOURCE << " requested fragment ";
			int fragment_id;
			MPI_Recv( &fragment_id, 1, MPI_INT, probe_status.MPI_SOURCE, DB_DATATOSEND_TAG, MPI_COMM_WORLD, MPI_STATUS_IGNORE );
			if( debug_msg ) CONT_LOG_MSG << fragment_id << " for MPI-copy" << endl;
			if (!sendFragsViaMPI( fragment_id, probe_status.MPI_SOURCE ))
				LOG_MSG << "***ERROR_2*** Could not send fragment " << fragment_id << " to Worker " << probe_status.MPI_SOURCE << endl;
			continue;
		}
		if( probe_status.MPI_TAG != BLAST_RETCODE_TAG ){
			LOG_MSG << "Unknown tag " << probe_status.MPI_TAG << " in writer loop!\n";
			MPI_Abort( MPI_COMM_WORLD, -1 );
		}
		// blast exited, git dem rezults frum dis hear node.
		int message_size = 0;
		MPI_Get_count( &probe_status, MPI_INT, &message_size);
		
		// receive a message of the form:
		// worker->writer:  result_id, frag_id, searched count, searched list, success count, success list, data size
		int* message_buf = (int*)malloc(message_size*sizeof(int));
		MPI_Recv( message_buf, message_size, MPI_INT, probe_status.MPI_SOURCE, 
				BLAST_RETCODE_TAG, MPI_COMM_WORLD, &status );

		int result_size = message_buf[ message_size - 1 ];
		if( debug_msg )
			LOG_MSG << "Receiving " << result_size << " bytes of results\n";
		// allocate space to receive these results
		ResultData_t rd;
		rd.notification = message_buf;
		meta_MPI_Alloc_mem( result_size, MPI_INFO_NULL, &rd.buf );
		rd.req = (MPI_Request*)malloc(sizeof(MPI_Request));
		memset(rd.req, 0, sizeof(MPI_Request));
		result_data.push_back( rd );
		// post a receive
		MPI_Irecv( rd.buf, result_size, MPI_BYTE, probe_status.MPI_SOURCE, 
			BLAST_RESULTS_TAG, MPI_COMM_WORLD, rd.req);

		// tell the worker to send results...
		int* command;
		meta_MPI_Alloc_mem( sizeof(int), MPI_INFO_NULL, &command );
		command[0] = message_buf[0];
		MPI_Request* req = (MPI_Request*)malloc( sizeof(MPI_Request) );
		memset( req, 0, sizeof(MPI_Request) );
		MPI_Isend( command, 1, MPI_INT, probe_status.MPI_SOURCE, ASSIGNMENT_TYPE, MPI_COMM_WORLD, req );
		addSendBuffer( command, req );

	}

	outputHTMLfooter();
	cleanupBLAST();
}

void MpiBlast::scheduler()
{

	MPI_Status status;
	int finished_nodes = 0;
	int current_queryI = 0;	// stores the index of the earliest query that hasn't been completely assigned

// initialize a FragmentDistribution which tracks the distribution of
// fragments on worker nodes
	FragmentDistribution frag_dist( total_frags, node_count );
// allocate the query-related data structures
//	BlastJob bj( total_frags, frag_dist );
//	PrecopySchedulerPolicy bj( total_frags, frag_dist );
	QueryData qd( NULL, NULL );
	query_data = vector< QueryData >( query_count, qd );
	for( int qI = 0; qI < query_data.size(); qI++ )
		query_data[ qI ].blast_job = new PrecopySchedulerPolicy( total_frags, frag_dist );
	output_queryI = 0;
	
// JA -- create DB access queue of workers waiting to access the database
// pair values are <worker rank, fragment id>
	queue< pair< int, int > > db_access_queue;

// JA -- number of workers currently copying DB fragments
//       value must always be less than concurrent_accesses
	int num_copying_db = 0;
	
	//
	// refresh each worker's local storage fragment list
	//
	for( uint workerI = 0; workerI < node_count; workerI++ ){
		if( workerI == scheduler_process || workerI == writer_process )
			continue;
		vector< int > fvec;
		RecvIntVec(fvec, workerI, DB_FRAGMENTS_TAG);
		set< int > fset = set< int >( fvec.begin(), fvec.end() );
		if( debug_msg ){
			LOG_MSG << "Node " << workerI;
			if( fvec.size() > 0 )
				CONT_LOG_MSG << " has fragments ";
			else
				CONT_LOG_MSG << " has no fragments";
			for( int fI = 0; fI < (int)fvec.size(); fI++ ){
				CONT_LOG_MSG << " " << fvec[ fI ];
			}
			CONT_LOG_MSG << endl;
		}
		// Update the FragmentDistribution
		frag_dist.update( workerI, fset );
		if( debug_msg ){
			LOG_MSG << "Received fragment list update from node " << workerI << endl;
		}
	}

	// tell the workers what fragments to query
	// ...
	while( true ){
		//
		// wait for a message from a worker
		//
		MPI_Status probe_status;
		MPI_Probe( MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &probe_status );
		if( debug_msg )
			LOG_MSG << "Scheduler got tag " << probe_status.MPI_TAG << " from " << probe_status.MPI_SOURCE << endl;
		
		if( probe_status.MPI_TAG == FRAGMENT_COPY_COMPLETE ){ // copy completed
			int new_fragment;
			MPI_Recv( &new_fragment, 1, MPI_INT, probe_status.MPI_SOURCE,
					FRAGMENT_COPY_COMPLETE, MPI_COMM_WORLD, &status );
				  
			frag_dist.copyCompleted( probe_status.MPI_SOURCE, new_fragment );
			
// db_access_queue should move into 'BlastJob'
// and be replaced by a more generic "assignment pending"
// queue that queues up any worker that gets assigned
// DO_NOTHING
			if (!db_access_queue.empty())
			{ // if not empty
				// JA -- Allow next worker in db_access_queue access to DB
				if (debug_msg){
					LOG_MSG << "Worker " << db_access_queue.front().first << " has turn for DB access now" << endl;
				}
				// construct a message and send it
				// the query id (index 3) doesn't matter for copy ops.
				int assign_array[4] = {COPY_FRAGMENT,0,db_access_queue.front().second,0};
				MPI_Send( assign_array, 4, MPI_INT, db_access_queue.front().first,
						ASSIGNMENT_TYPE, MPI_COMM_WORLD );
				db_access_queue.pop();  // remove from top of Queue
			} else {
				num_copying_db--;
			}
		} else if( probe_status.MPI_TAG == WORKER_IDLE ){
			// this worker's bored, give it something to do...
			int idle_code;

			MPI_Recv( &idle_code, 1, MPI_INT, probe_status.MPI_SOURCE, WORKER_IDLE, MPI_COMM_WORLD, &status );

/*	LBC -- only 1 send operation to make an assignment
	assign_array[0] = assignment
	assign_array[1] = db-name as int
	assign_array[2] = fragment_id to copy/search
	assign_array[3] .. [N] = search these queries
	In the future, multiple databases: [2]=#of frags, [3-N]=frag#s, [N+1-M]=query#s
	Not sure how multiple queries will be implemented. 
	Mechanism is in place in worker to recieve an arbitrarily long assignment_array
		where [3]->[size-1] are queries

*/
		
			int assign_array[4];
			assign_array[0] = DO_NOTHING;
			
			// if current_queryI == query_count then we have assigned all the work already
			// just tell this worker to quit
			
			// loop until we get an assignment or until we 
			// exhaust the list of unsearched queries
			int queryI = current_queryI;
			while( queryI < query_count && (assign_array[0] == DO_NOTHING || assign_array[0] == SEARCH_COMPLETE) ){
				query_data[ queryI ].blast_job->getAssignment( probe_status.MPI_SOURCE, assign_array[0], assign_array[2] );
				// if current_queryI has been completely assigned, advance it to the next query
				if( queryI == current_queryI && assign_array[0] == SEARCH_COMPLETE )
					current_queryI++;
				queryI++;
				if( queryI == query_count )
					break;
			}
			queryI--;  // it was incremented before breaking the loop, decrement it
			assign_array[3] = queryI;

			if( current_queryI == query_count )//all done.  worker can quit now...
				assign_array[0] = WORKER_QUIT;
				
			if( debug_msg )
				LOG_MSG << "Giving " << probe_status.MPI_SOURCE << " assignment " << assign_array[0] << " on frag " << assign_array[2] << endl;
			
			// if we're copying, this message may get used as a synchronization point
			// for DB pipelining
			if( assign_array[0] != COPY_FRAGMENT )
				MPI_Send( &assign_array, 4, MPI_INT, probe_status.MPI_SOURCE, ASSIGNMENT_TYPE, MPI_COMM_WORLD );

			// JA -- if assignment is search, send message now so that Worker can start
		    //       if assignment is copy, place worker in Q to serialize access to DB
	        //       if search is complete, increment finished node count
			if (assign_array[0] == SEARCH_FRAGMENT){
			}else if (assign_array[0] == COPY_FRAGMENT){
				if (num_copying_db >= concurrent_accesses && copy_via != COPY_VIA_NONE){
					// JA -- put on Queue and wait turn for DB access
					db_access_queue.push( pair< int, int >( probe_status.MPI_SOURCE, assign_array[2] ) );
					if (debug_msg)
						LOG_MSG << "Node " << probe_status.MPI_SOURCE << " Waiting turn for DB access" << endl;
				} else {
					num_copying_db++;
					// JA -- send message so as to start copying
					if (debug_msg)
						LOG_MSG << "Node " << probe_status.MPI_SOURCE << " turn for DB access is now" << endl;
					MPI_Send( &assign_array, 4, MPI_INT, probe_status.MPI_SOURCE, ASSIGNMENT_TYPE, MPI_COMM_WORLD );
				}
         	} else if( assign_array[0] == WORKER_QUIT ){
	         	finished_nodes++;
 	           	if( finished_nodes == node_count - 2 )
	   	         	break;
				else
            		continue;
      		}
		} else {
			// Catch any out of place messages
			LOG_MSG << "Error: Received msg " << probe_status.MPI_TAG << " from " << probe_status.MPI_SOURCE << endl;
			throw __FILE__ "scheduler(): Illegal message in main loop!";
		}
	} // end while (true)

}

/*
 * quickly check whether the writer wants results
 */
void MpiBlast::checkResultRequests()
{
	MPI_Status status;
	int flag = 0;
	MPI_Iprobe( writer_process, ASSIGNMENT_TYPE, MPI_COMM_WORLD, &flag, &status );
	if( flag ){
		// wait for previous sends to complete before sending this one
		waitForSendCompletions();
		// the writer wants this worker to send results
		int send_id = 0;
		MPI_Recv( &send_id, 1, MPI_INT, writer_process, ASSIGNMENT_TYPE, MPI_COMM_WORLD, &status );
		// send the result asynchronously
		map< int, std::pair< int, unsigned char* > >::iterator result_iter = result_id_data_map.find( send_id );
		if( result_iter == result_id_data_map.end() ){
			LOG_MSG << "Error.  unable to find results with id: " << send_id << endl;
			MPI_Abort( MPI_COMM_WORLD, -1 );
		}
		unsigned char* send_data = result_iter->second.second;
		int send_size = result_iter->second.first;
		if( debug_msg )
			LOG_MSG << "Sending " << send_size << " bytes in result id: " << result_iter->first << endl;
// async send
		MPI_Request* req = (MPI_Request*)malloc( sizeof( MPI_Request ) );
		MPI_Irsend( send_data, send_size, MPI_BYTE, writer_process, BLAST_RESULTS_TAG, MPI_COMM_WORLD, req );
		addSendBuffer( send_data, req );

// sync send
//		MPI_Send( send_data, send_size, MPI_BYTE, writer_process, BLAST_RESULTS_TAG, MPI_COMM_WORLD );
		result_id_data_map.erase( result_iter );
	}
}

void MpiBlast::worker(){
	
	meta_MPE_Log_event(worker_file_setup_start, 0,"setting up worker files");
	//
	// create blast alias file
	//    
	string alias_basename = config.localPath() + database_name + "XXXXXX";
	getTempFileName( alias_basename );
	string alias_filename = alias_basename + "." + db_type + "al";
	moveFile( alias_basename, alias_filename );
	// be sure the alias file gets deleted before exiting
	registerFileToDelete( alias_filename );

	stringstream rank_str;
	rank_str << rank;

#ifdef WIN32
	const char* bitbucket = "NUL";
#else
	const char* bitbucket = "/dev/null";
#endif

	meta_MPE_Log_event(worker_file_setup_end, 0,"done setting up worker files");

	//
	// initialize the NCBI Toolbox, create blast command line
	//
	meta_MPE_Log_event(init_ncbi_start, 0,"start initNCBI()");

	addOpt( worker_opts, 'i', query_filename.c_str() );

	/** are we using a gi filter? */
	if ( filter_filename != "" )
	        addOpt( worker_opts, 'l', filter_filename.c_str() );

	if( use_binary_asn )
		addOpt( worker_opts, 'm', "11" );	/** ask for binary SeqAlign output */
	else
		addOpt( worker_opts, 'm', "10" );	/** ask for text SeqAlign output */
	addOpt( worker_opts, 'o', bitbucket );	/** since results get written to memory we can use /dev/null for any other output */
	addOpt( worker_opts, 'd', alias_basename.c_str() );

	// Init the NCBI library with the contrived blast command line
	running_blast = true;
	initNCBI( worker_opts );
	running_blast = false;
		
	meta_MPE_Log_event(init_ncbi_end, 0,"end initNCBI()");
	
	MPI_Status status;
	
	// These variables manage the database fragment list (for sending, receiving and writing
	// to disk).
	vector<int> fragment_list;
	int frag_id;	/**< The fragment currently being searched */
	
	// construct the fragment list file name
	string fragment_filename = config.localPath() + database_name + FRAG_LIST_EXTENSION;	

	int query_id;	/**< The query that is currently being worked on */
	bool first_search=true;

	//
	// tell the scheduler node what fragments this node has	
	//
	meta_MPE_Log_event(send_frag_list_start, 0,"begin send frag list to scheduler");

	frag_list_file = FragmentListFile( fragment_filename, config, database_name, db_type );
	frag_list_file.setTimestamps( all_dates, total_datelen );
	frag_list_file.checkTimestamps();
	frag_list_file.SendList( scheduler_process );		
	if( debug_msg ){
		LOG_MSG << "Fragment list sent." << endl;
	}
	meta_MPE_Log_event(send_frag_list_end, 0,"end send frag list to scheduler");

	// loop until the assignment is SEARCH_COMPLETE
	int* assign_array = NULL;	/**< will store the assignment */
	while( true ){

		//
		// tell scheduler we are idle
		//
		int idle_msg = WORKER_IDLE;
   		MPI_Send( &idle_msg, 1, MPI_INT, scheduler_process, WORKER_IDLE, MPI_COMM_WORLD );
  		if( debug_msg ){
     			LOG_MSG << "Idle message sent" << endl;
  		}
		
		// quickly check whether the writer wants results
		checkResultRequests();

		//
		// get an assignment from the scheduler
		//
		int array_size;
		memset( &status, 0, sizeof(MPI_Status) );
		MPI_Probe( scheduler_process, ASSIGNMENT_TYPE, MPI_COMM_WORLD, &status );
		MPI_Get_count( &status, MPI_INT, &array_size );
		assign_array = (int*)malloc(sizeof(int)*array_size);
		
		MPI_Recv( assign_array, array_size, MPI_INT, scheduler_process, ASSIGNMENT_TYPE, MPI_COMM_WORLD, &status );
		query_id = assign_array[3];
		frag_id = assign_array[2];

		if( debug_msg ){
    			LOG_MSG << "Assignment " << assign_array[0] << " is " << array_size << " ints long.\n";
   		}


		//
		// we have an assignment, now act on it!
		//
		// exit the worker loop if there's no more work to do
		if( assign_array[0] == SEARCH_COMPLETE || assign_array[0] == WORKER_QUIT )
			break;	// don't delete assign_array yet!

		if (assign_array[0] != DO_NOTHING ) {
			// JA -- If assignment is SEARCH_FRAGMENT, worker will immediately receive these messages
			//       If assignment is COPY_FRAGMENT, worker is put in queue and will receive messages when at front of queue
		
			if( assign_array[0] == COPY_FRAGMENT ) {
				// first check whether another worker has already copied this
				// fragment to the local storage directory
				// note that the scheduler mustn't allow two workers to copy
				// the same fragment simultaneously otherwise they could overwrite
				// each other's files when multiple workers share local storage
				frag_list_file = FragmentListFile( fragment_filename, config, database_name, db_type );
				frag_list_file.setTimestamps( all_dates, total_datelen );
				frag_list_file.checkTimestamps();
				if( !frag_list_file.contains( frag_id ) ) {
					// copy the database fragments 
					// .nhr, .nin, .nsq, .nnd, .nni, .nsd, .nsi for nucleotide
					// .phr, .pin, .psq, .pnd, .pni, .psd, .psi for amino acids
					bool copy_success = true;
					const vector<string>& extensions = FragmentExtensions(db_type);
				
					char fragid[8];
					memset( fragid, 0, 8 );
					// always support 3 digit fragment identifiers
					sprintf( fragid, "%03d", frag_id );
					if( debug_msg ){
						LOG_MSG << "copying fragment " << fragid << endl;
					}

					meta_MPE_Log_event(cpstart,0,"start copy");
					if (copy_via != COPY_VIA_MPI) { // copy via cp/rcp/scp
						for( uint extI = 0; extI < extensions.size(); extI++ ){
							string copy_fragment_name = database_name + ".";
							copy_fragment_name += fragid + extensions[ extI ];
							string copy_src = config.sharedPath() + copy_fragment_name;
							string copy_dest = config.localPath() + copy_fragment_name;
							if( copyFile( copy_src, copy_dest, copy_via ) != 0 )
								copy_success = false;
						}
					} else { // copy via mpi
						copy_success = recvFragsViaMPI( frag_id );
					}	
					meta_MPE_Log_event(cpend,0,"end copy");

					if( copy_success ){
						frag_list_file.addFragment( frag_id );
					}
					else{
						LOG_MSG << "(" << rank << ") unable to copy fragment!" << endl;
						MPI_Abort( MPI_COMM_WORLD, -1 );
					}
				}	// end "if( !fragment_list_file.contains( frag_id ) )"

				//
				// Tell the scheduler that the fragment copy completed successfully
				//
				MPI_Send(&frag_id, 1, MPI_INT, scheduler_process, FRAGMENT_COPY_COMPLETE, MPI_COMM_WORLD );	
				delete[] assign_array;
				continue;
			} // end "if (assignment == COPY_FRAGMENT)"

			// Assignment is SEARCH_FRAGMENT
			// fill alias file with the names of fragments to search.
			//    
			if( debug_msg )
				LOG_MSG << "Searching query " << assign_array[3] << " vs frag " << assign_array[2] << endl;
			vector<int> single_frag_vector(1, frag_id);
			WriteAliasFile(alias_filename, single_frag_vector);
		
			//
			// initialize BLAST with the current data set
			//
			meta_MPE_Log_event(init_blast_start, 0,"initBLAST() start");
			running_blast = true;
			int retcode = initBLAST();
			running_blast = false;
			meta_MPE_Log_event(init_blast_end, 0,"initBLAST() end");
			if( retcode != 0 ) {
				// initBLAST errored out.  bail out of the search
				LOG_MSG << "Error: got initBLAST exit code " << retcode << endl;
				throw __FILE__ "(MpiBlast::worker): Received non-zero initBLAST exit code ";
			}
		
			
			//
			// If this is the first search, load the query sequences
			//
			if( first_search ){
				meta_MPE_Log_event(load_queries_start,0,"start loadQueries()");
				loadQueries();
				first_search = false;
				meta_MPE_Log_event(load_queries_end,0,"end loadQueries()");
			}

			//
			// execute blast
			//
			if( debug_msg ){
				LOG_MSG << "node " << rank << " Executing BLAST search\n";
			}
			meta_MPE_Log_event(blaststart,0,"start blastall");
			running_blast = true;
			retcode = runBLAST( SEARCH_MODE, query_id, query_id );
			running_blast = false;
			meta_MPE_Log_event(blastend,0,"end blastall");
			if( retcode != 0 ){
				LOG_MSG << "blastall exited with code: " << retcode << endl;
				MPI_Abort( MPI_COMM_WORLD, -1 );
			}


			meta_MPE_Log_event(cleanup_blast_start,0,"cleanupBLAST() start");
			running_blast = true;
			cleanupBLAST();
			running_blast = false;
			meta_MPE_Log_event(cleanup_blast_end,0,"cleanupBLAST() end");

			if( debug_msg ){
				LOG_MSG << "node " << rank << " blast retcode= " << retcode << endl;
			}
	

			//
			// send writer process notification that blast has completed
			// the writer will request the results data when it is ready
			//
			meta_MPE_Log_event(send_results_start,0,"sendResults() start");
			vector< int > searched_ids = vector< int >( 1, query_id );
			notifyWriterOfCompletion( writer_process, frag_id, searched_ids );
			meta_MPE_Log_event(send_results_end,0,"sendResults() end");
		}else{
			// assignment was DO_NOTHING...
			// in the future the worker shouldn't ever see this message.
			// for now, just have it sleep briefly to avoid swamping the
			// scheduler with requests
#ifndef WIN32
			sleep(1);
#else
			Sleep(1000);
#endif
		}
		free(assign_array);

	} // end while (true) worker loop

	// search is now complete
	// finish any pending send operations
	waitForSendCompletions();

	// free query sequence memory
	cleanupQueries();
	
	// wait for the writer to request all remaining results
	while( result_id_data_map.size() > 0 )
	{
		checkResultRequests();
#ifdef WIN32
		Sleep(50);
#else
		usleep(50000);
#endif
	}

	while(assign_array[0] != WORKER_QUIT){
		int array_size = 0;
		MPI_Probe( scheduler_process, ASSIGNMENT_TYPE, MPI_COMM_WORLD, &status );
		MPI_Get_count( &status, MPI_INT, &array_size );
		free(assign_array);
		assign_array = (int*)malloc(sizeof(int)*array_size);
		MPI_Recv( assign_array, array_size, MPI_INT, scheduler_process, ASSIGNMENT_TYPE, MPI_COMM_WORLD, &status );
	}
	free(assign_array);

	//  DEBUG
	MpiBlastDisableBioseqFetch();

}

void  MpiBlast::WriteAliasFile( const string& alias_filename, const vector< int >& fragment_list)
{
	if(fragment_list.size() == 0){
		return;	// don't write a file
		// this isn't an error condition when # workers > # db frags
//		throw __FILE__ "(WriteAliasFile) Empty fragment_list";
	}
	
	ofstream alias_file( alias_filename.c_str() );

	if( !alias_file.is_open() ){
		LOG_MSG << "Error opening " << alias_filename << endl;
		throw __FILE__ "(MpiBlast::worker): Unable to open alias file";
	}

	alias_file << "TITLE " << config.localPath() << database_name << endl;
	alias_file << "DBLIST";
	
	for( uint iter = 0; iter != fragment_list.size(); iter++ ){
		alias_file << " " << database_name << ".";
		char fragid[8];
		
		memset(fragid, 0, 8);

		// always use 3 digit fragment identifiers
		sprintf( fragid, "%03d", fragment_list[ iter ] );

		alias_file << fragid;
		if( debug_msg ){
			LOG_MSG << "Will search "<<database_name << "." << fragid << endl;
		}
	}
	alias_file << endl;
	alias_file.close();
		
}

// need getpid for debugging under win32
//#include <process.h>

int MpiBlast::main( int argc, char* argv[] )
{
#if defined(WIN32) && defined(_DEBUG)
	// for debugging under win32, wait so the process can be attached.
//	LOG_MSG << "Process id: " << _getpid() << endl;
	cerr << "Gentlemen, start your debuggers!\n";
	Sleep( 7000 );
#endif
#ifdef WIN32
	SetConsoleCtrlHandler( (PHANDLER_ROUTINE) Win32CtrlHandler, TRUE );
#endif
	if( argc == 0 ){
		cerr << "Incorrect usage\n";
		MPI_Abort( MPI_COMM_WORLD, -1 );
		throw __FILE__ "(MpiBlast::main): Incorrect usage";
	}
	exec_path = argv[0];
	exec_path = getPath( exec_path );

	//
	// must read -d, -i, and -o options
	//
	const char* short_opts = "-d:i:o:p:m:O:l:";
	int ac = argc;
	char** av = argv;
	int opt_code, opt_type = 0;
	opterr = 0;
	string log_file_name;
//	string config_f_name = MpiBlastConfig::defaultConfigFileName();
	string config_f_name = ".ncbirc";
	bool copy_via_set = false;
	int config_opt = 0;
	int long_index = 0;
	int niceness = 0;
	bool lock_set = false;

	writer_opts.push_back( "blastall" );
	scheduler_opts.push_back( "blastall" );
	worker_opts.push_back( "blastall" );
	struct option long_opts[] = {
		{"config-file", required_argument, &config_opt, 1 },
		{"debug", optional_argument, &config_opt, 2 },
		{"pro-phile", required_argument, &config_opt, 3 },
		{"removedb", no_argument, &config_opt, 4 },
		{"version", no_argument, &config_opt, 5 },
		{"disable-mpi-db", no_argument, &config_opt, 6 },
		{"nice", required_argument, &config_opt, 7 },
		{"concurrent", required_argument, &config_opt, 8},
		{"mpi-size", required_argument, &config_opt, 9},
		{"copy-via", required_argument, &config_opt, 10},
		{"lock", optional_argument, &config_opt, 11},
		{"resume-run", no_argument, &config_opt, 12},
		{"scheduler-rank", required_argument, &config_opt, 13},
		{"altschul-reference", no_argument, &config_opt, 14},
		{"db-replicate-count", required_argument, &config_opt, 15},
		{0,0,0,0}	// signifies termination of option list
	};
	
	while((opt_code = getopt_long( ac, av, short_opts, long_opts, &long_index ) ) != EOF ){
		switch( opt_code ){
			case 0:
				switch( config_opt ){
					case 1:	// --config-file
						config_f_name = optarg;
						break;
					case 2:	// --debug
						if( optarg != NULL )
							log_file_name = optarg;
						debug = true;
						debug_msg = true;
						debug_bsfetch = true;
						debug_bslookup = true;
						break;
					case 3:	// was pro-phile -- now deprecated
						cerr << "WARNING: --pro-phile is no longer supported\n";
						break;
					case 4:	// --removedb
						remove_db = true;
						break;
					case 5:	// --version
						cout << PACKAGE << " version " << VERSION << endl;
						break;
					case 6:	// --disable-mpi-db
						disable_mpi_db = true;
						break;
					case 7:	// --nice
						niceness = atoi(optarg);
						break;
					case 8:	// --concurrent
						concurrent_accesses = atoi(optarg);
						if( concurrent_accesses > 1 && !lock_set )
							lock_mbf = true;
						break;
					case 9:	// --mpi-size
						max_data_to_send = (uint64) atoi(optarg);
						break;
					case 10:	// --copy-via
						//options = cp, rcp, scp, none/pvfs/gpfs, mpi
						copy_via_set = true;
						if (!strcmp(optarg,"cp")){
							copy_via = COPY_VIA_CP;
						} else if (!strcmp(optarg,"rcp")){
							copy_via = COPY_VIA_RCP;
						} else if (!strcmp(optarg, "scp")){
							copy_via = COPY_VIA_SCP;
						} else if (!strcmp(optarg, "none")){
							copy_via = COPY_VIA_NONE;
						} else {
							copy_via = COPY_VIA_MPI;
						}
						break;
					case 11:	// --lock
						lock_mbf = true;
						if( optarg != NULL && strcmp( optarg, "off" ) )
							lock_mbf = false;
						lock_set = true;
						break;
					case 12:	// --resume-run
						resume_run = true;
						break;
					case 13:	// --scheduler-rank
						scheduler_process=atoi(optarg);
						break;
					case 14:	// --altschul-reference
						strict_output_conformance = TRUE;
						break;
					case 15:	// --db-replicate-count
						PrecopySchedulerPolicy::db_copy_number = atoi(optarg);
						break;
				}	
				break;

			case 'i':
				query_filename = optarg;
				addOpt( writer_opts, opt_code, optarg );
				break;
			case 'o':
				output_filename = optarg;
				addOpt( writer_opts, opt_code, optarg );
				break;
			case 'd':
				database_name = optarg;
				break;
			case 'p':
				addOpt( writer_opts, opt_code, optarg );
				addOpt( worker_opts, opt_code, optarg );
				blast_type = optarg;
				break;
			case 'm':
				// worker always uses -m 11
				output_format = atoi( optarg );
				addOpt( writer_opts, opt_code, optarg );
				break;
			case 'O':
				// Only the master is allowed to see this
				// flag since workers must use -m 11. Don't let the
				// -O force ASCII ASN.1 output on the workers! 
				addOpt( writer_opts, opt_code, optarg );
				break;
			case 'l':
				// this specifies a gi list to use when searching
				// for now just assume it's located on shared storage
				filter_filename = optarg;
				addOpt( writer_opts, opt_code, optarg );
				break;
			case 1:
				worker_opts.push_back( optarg );
				writer_opts.push_back( optarg );
				break;
			case '?':
				addOpt( worker_opts, optopt, optarg );
				addOpt( writer_opts, optopt, optarg );
				opt_type = optopt;
				break;
			default:
				addOpt( worker_opts, opt_type, optarg );
				addOpt( writer_opts, opt_type, optarg );
				opt_type = optopt;
				break;
		}
	}

#ifndef WIN32
	//Set the niceness if given as cmd line arg & we're in unix
	if (niceness > 0)
		setpriority(PRIO_PROCESS, 0, niceness);
#endif

	try{
		config = MpiBlastConfig( config_f_name );
		
		// if the user wants a "clean" run, create a temp
		// directory on the local storage device for the job
		if( remove_db )
			config.createLocalTempDir();

		localPath = config.localPath();
		if (config.localPath() == config.sharedPath() ) {
			// if the user didn't specify copy_via and the config
			// file has identical shared and local then default
			// to --copy_via=none for ease of use
			if( !copy_via_set )
				copy_via = COPY_VIA_NONE;
			if( copy_via != COPY_VIA_NONE ){
				// can't have same shared/local and not have copy_via=none
				cerr << "Error: You must set --copy_via=none when Shared and Local storage are identical\n";
				terminateProgram(1);
			}
			sameSharedLocal = 1;
		} else {
			if ( copy_via == COPY_VIA_NONE){
				// can't have different shared/local when copying via none
                                cerr << "Error: Shared and Local storage must be identical when --copy_via=none\n";
                                terminateProgram(1);
                        }
			sameSharedLocal = 0;
		}

		db_type = "n"; //Default to nucleotide
		if( blast_type == "blastp" || blast_type == "blastx" ){
			db_type = "p";
		}
				
		// set location of substitution matrices if not set by the user
		if( getenv( "BLASTMAT" ) == NULL ){
			string env_str = "BLASTMAT=" + config.sharedPath();
			setEnvironmentVariable( env_str.c_str() );
		}
		
		//
		// open the log file
		//
		if( log_file_name != "" ){
			ostringstream log_file_str;
			log_file_str << log_file_name << "." << rank;
			log_file.open( log_file_str.str().c_str() );
			
			if( !log_file.is_open() ){
				cerr << "Error opening log file \"" << log_file_name << "\"\n";
				throw __FILE__ "(main): Unable to open log file";
			}
			
			if( debug_msg ){
				cerr << "logging to " << log_file_str.str() << endl;
			}
			
			log_stream = &log_file;
		}

		//
		// add database name for master
		//
		// The old version
		// string shared_db_name = config.sharedPath() + database_name;
		// Assume that the database resides on either the master, or a valid nfs mount
		string shared_db_name;
		if( !isRemotePath( config.sharedPath() ) ){
			// NFS mount
			shared_db_name = config.sharedPath() + database_name;
		}
		else{
			// On master
			string::size_type pos = config.sharedPath().find(":", 0);
			pos ++;
			shared_db_name = config.sharedPath().substr(pos, config.sharedPath().length() - pos)
				+ database_name;
		}
		
		// have rank 0 count the number of fragments 
		// and send the number to workers
		if( rank == 0 ){
			string mbf_filename = shared_db_name + ".mbf";
			FragmentListFile frag_list_file( mbf_filename, config, database_name, db_type );
			total_frags = frag_list_file.fragmentCount();
			MPI_Bcast( &total_frags, 1, MPI_UNSIGNED, 0, MPI_COMM_WORLD );
		}else{
			MPI_Bcast( &total_frags, 1, MPI_UNSIGNED, 0, MPI_COMM_WORLD );
		}

		addOpt( writer_opts, 'd', shared_db_name.c_str() );
		
		query_file = getFilename( query_filename );
		local_query_filename = config.localPath() + query_file;

		// for gi list filters:
		filter_file = getFilename( filter_filename );
		local_filter_filename = config.localPath() + filter_file;

		//read the output file, and return the number of the last query w/output -- X
		//read off the first X queries, and place the remaining ones into a new query file
		string resume_data_filename = query_filename + "." + database_name + ".resumedata";
		if( rank == 0 )
		{
			registerFileToDelete( resume_data_filename );
			if(resume_run){
				if(!doesFileExist(output_filename) || (output_format == 7) ||(output_format == 8) || (output_format == 10) || (output_format == 11) ){
					LOG_MSG << "To resume a run the outputfile must already exist\n";
					LOG_MSG << "Only -T and -m 1-6 & 9 are supported for --resume\n";
					terminateProgram(-1);	
				}

				unsigned int lastquerynum = determineLastQueryNum(output_filename);
				string new_qf = config.localPath() + query_file + "XXXXXX";
				makeNewQuery(new_qf,query_filename, lastquerynum);
				// Overwrite the query value in the writer's argument list
				addOpt( writer_opts, 'i', new_qf.c_str() );
				registerFileToDelete( new_qf );
				query_filename = new_qf ; 
			}
		}

		// Have the rank 0 process send the query and gi filter file to all the other processes
		if( rank == 0 ){

			if( !isRemotePath( query_filename ) || copy_via == COPY_VIA_MPI){
				// The query is either (a) on the masters local disk or (b) visible to the master
				// via an NFS mount. No path name adjustment is required.
				// JA -- or (c) when using MPI to copy, rank 0 must have direct access to query
			}
			else{
				// The query resides on another host (not local to the master). Use rcp to copy the query to the
				// masters local disk.
				if( copyFile( query_filename, local_query_filename, copy_via ) != 0 ){
					LOG_MSG << "Unable to copy " << query_filename << " to "
					     << local_query_filename << endl;
					     
					throw __FILE__ "(main): Unable to copy query to master";
				}

				// Make sure the master now reads from the local copy
				query_filename = local_query_filename;

				// Overwrite the query value in the writer's argument list
				addOpt( writer_opts, 'i', local_query_filename.c_str() );
				registerFileToDelete( local_query_filename );
			}

			//
			// broadcast the query to every node
			//
			meta_MPE_Log_event(bcast_query_start,0,"broadcast query start");
			broadcastFile( query_filename );
			meta_MPE_Log_event(bcast_query_end,0,"broadcast query end");

			// filter file handling
			if ( filter_filename != "" ){
				// MC - This is crudely cut and pasted from the query file broadcast.
				// It ought to be unified and refactored
				// Have the rank 0 process send the filter file to all the other processes
				if( !isRemotePath( filter_filename ) || copy_via == COPY_VIA_MPI ){
					// The filter is either (a) on the masters local disk or (b) visible to the master
					// via an NFS mount. No path name adjustment is required.
					// JA -- or (c) when using MPI to copy, rank 0 must have direct access to filter
				}
				else{
					// The filter resides on another host (not local to the master). 
					// Use rcp to copy the filter to the
					// masters local disk.
					if( copyFile( filter_filename, local_filter_filename, copy_via ) != 0 ){
						LOG_MSG << "Unable to copy " << filter_filename << " to "
							<< local_filter_filename << endl;

						throw __FILE__ "(main): Unable to copy filter to master";
					}

					// Make sure the master now reads from the local copy
					filter_filename = local_filter_filename;

					// Overwrite the query value in the writer's argument list
					addOpt( writer_opts, 'l', local_filter_filename.c_str() );
					registerFileToDelete( local_filter_filename );
				}

				//
				// broadcast the filter to every node
				//
				meta_MPE_Log_event(bcast_filter_start,0,"broadcast filter start");
				broadcastFile( filter_filename );
				meta_MPE_Log_event(bcast_query_end,0,"broadcast filter end");
			}



			meta_MPE_Log_event(init_ncbi_start, 0,"start initNCBI()");

			running_blast = true;
			initNCBI( writer_opts );
			int rval = initBLAST();
			running_blast = false;

			meta_MPE_Log_event(init_ncbi_end, 0,"end initNCBI()");

			if( rval != 0 ){
				// bail out on error for now.
				LOG_MSG << "Error: got initBLAST exit code " << rval << endl;
				throw __FILE__ "(MpiBlast::master): Received non-zero initBLAST exit code";
			}

			if( debug_msg ) {
				LOG_MSG << "Init blast error code " << rval << endl;
			}
	
		// load the queries and count how many there are
			meta_MPE_Log_event(load_queries_start,0,"start loadQueries()");
			query_count = loadQueries();
			meta_MPE_Log_event(load_queries_start,0,"start loadQueries()");

		// broadcast the array size
			MPI_Bcast( &query_count, 1, MPI_INT, 0, MPI_COMM_WORLD );

		// allocate space for search statistics
			stats_array = (Int8**)MemNew( sizeof(Int8*)*query_count );
			stats_array[0] = (Int8*)MemNew( sizeof(Int8)*12*query_count );
			memset( stats_array[0], 0, sizeof(Int8)*12*query_count );
			for( int qI = 0; qI < query_count; qI++ )
				stats_array[qI] = &stats_array[0][qI*12];

		// allocate the eff. query and db length arrays
			options->query_adjustments = (Int8*)MemNew( sizeof( Int8 ) * query_count );
			options->effective_db_lengths = (Int8*)MemNew( sizeof( Int8 ) * query_count );
		
		// if resuming, read the query adjustments and effective database lengths
			if( resume_run && doesFileExist( resume_data_filename ) )
			{
				ifstream resume_file( resume_data_filename.c_str() );
				if( !resume_file.is_open() )
				{
					cerr << "Error opening \"" << resume_data_filename << "\"\n";
					return -1;
				}
				if( debug_msg )
					LOG_MSG << "Reading effective search space values for each query from the resume file " << resume_data_filename << endl;
				vector< Int8 > resume_data;
				Int8 resume_val;
				while( resume_file >> resume_val )
					resume_data.push_back( resume_val );

				int rI = resume_data.size() - query_count * 2;
				if( rI < 0 )
				{
					cerr << "Error reading resume data file: " << resume_data_filename << endl;
					cerr << "Please delete the file and try again\n";
					return -1;
				}
				for( int qI = 0; qI < query_count; qI++ )
				{
					options->query_adjustments[qI] = resume_data[rI++];
					options->effective_db_lengths[qI] = resume_data[rI++];
				}
			}else
			{
			// get the query adjustments and effective database lengths for all queries
				runBLAST( COLLECT_STATS_MODE, 0, query_count - 1 );
			// write the query adjustments to a file that can be read if the search
			// needs to be resumed, or deleted when the search finishes
				ofstream resume_file( resume_data_filename.c_str() );
				if( resume_file.is_open() )
					for( int qI = 0; qI < query_count; qI++ )
						resume_file << options->query_adjustments[qI] << '\t' << options->effective_db_lengths[qI] << endl;
			}
	
		// send query adjustments and database lengths to workers

			meta_MPE_Log_event(bcast_qadj_start,0,"bcast query adjustments start");
			MPI_Bcast( options->query_adjustments, query_count, MPI_LONG_LONG_INT, 0, MPI_COMM_WORLD );

			MPI_Bcast( options->effective_db_lengths, query_count, MPI_LONG_LONG_INT, 0, MPI_COMM_WORLD );
			meta_MPE_Log_event(bcast_qadj_end,0,"bcast query adjustments end");

		}else{
			//
			// receive the query file broadcast
			//
			meta_MPE_Log_event(bcast_query_start,0,"broadcast query start");
			// create local query tempfile
			query_filename = local_query_filename + "XXXXXX";
			getTempFileName( query_filename );
			// receive the query file broadcast
			recvBroadcastFile( query_filename );
			// make sure the query file gets deleted before exiting
			registerFileToDelete( query_filename );
			if( debug_msg ){
				LOG_MSG << "Query file received as " << local_query_filename << endl;
			}
			meta_MPE_Log_event(bcast_query_end,0,"broadcast query end");

			if ( filter_filename != "" ){
				//
				// receive the gi filter file broadcast
				//
				meta_MPE_Log_event(bcast_filter_start,0,"broadcast filter start");
				// create local filter tempfile
				filter_filename = local_filter_filename + "XXXXXX";
				getTempFileName( filter_filename );
				// receive the filter file broadcast
				recvBroadcastFile( filter_filename );
				// make sure the filter file gets deleted before exiting
				registerFileToDelete( filter_filename );
				if( debug_msg ){
						LOG_MSG << "Filter file received as " << local_filter_filename << endl;
				}
				meta_MPE_Log_event(bcast_filter_end,0,"broadcast filter end");
			}

			//
			// receive the query length adjustments and effective db sizes from the writer
			//
			if( debug_msg ){
				LOG_MSG << "Receiving query length adjustments and effective db sizes from the master\n";
			}
			// receive the array size
			MPI_Bcast( &query_count, 1, MPI_INT, 0, MPI_COMM_WORLD );

			// allocate space for search statistics
			stats_array = (Int8**)MemNew( sizeof(Int8*)*query_count );
			stats_array[0] = (Int8*)MemNew( sizeof(Int8)*12*query_count );
			memset( stats_array[0], 0, sizeof(Int8)*12*query_count );
			for( int qI = 0; qI < query_count; qI++ )
				stats_array[qI] = &stats_array[0][qI*12];

			// allocate space and receive the array
			query_adj_array = (Int8*)MemNew( sizeof( Int8 ) * query_count );
			db_len_array = (Int8*)MemNew( sizeof( Int8 ) * query_count );


			meta_MPE_Log_event(bcast_qadj_start,0,"bcast query adjustments start");
			MPI_Bcast( query_adj_array, query_count, MPI_LONG_LONG_INT, 0, MPI_COMM_WORLD );
			MPI_Bcast( db_len_array, query_count, MPI_LONG_LONG_INT, 0, MPI_COMM_WORLD );
			meta_MPE_Log_event(bcast_qadj_end,0,"bcast query adjustments end");

			if( debug_msg ){
				LOG_MSG << query_count << " adjustments and effective db sizes received successfully\n";
			}
		}
		
		//
		// have rank 0 send database fragment timestamps to all nodes
		//
		if( rank == 0 ){
			// open the database info files
			char* m_dbname = (char*)malloc( shared_db_name.size() + 1 );
			strcpy( m_dbname, shared_db_name.c_str() );
			ReadDBFILEPtr rdfp = readdb_new_ex2( m_dbname, db_type == "p", READDB_NEW_DO_REPORT, NULL, NULL );
			free( m_dbname );
			ReadDBFILEPtr cur_rdfp = rdfp;
			total_datelen = 0;
			while( cur_rdfp != NULL ){
				char* date_str = readdb_get_date( cur_rdfp );
				total_datelen += strlen( date_str ) + 1;
				cur_rdfp = cur_rdfp->next;
			}
			all_dates = (char*)malloc( total_datelen );
			char* cur_date = all_dates;
			cur_rdfp = rdfp;
			while( cur_rdfp != NULL ){
				char* date_str = readdb_get_date( cur_rdfp );
				strcpy( cur_date, date_str );
				cur_date += strlen( date_str ) + 1;
				cur_rdfp = cur_rdfp->next;
			}
			MPI_Bcast( &total_datelen, 1, MPI_INT, 0, MPI_COMM_WORLD );
			MPI_Bcast( all_dates, total_datelen, MPI_CHAR, 0, MPI_COMM_WORLD );
			rdfp = readdb_destruct( rdfp );
			if( debug_msg )	LOG_MSG << "First date: " << all_dates << endl;
		}else{
			MPI_Bcast( &total_datelen, 1, MPI_INT, 0, MPI_COMM_WORLD );
			all_dates = (char*)malloc( total_datelen );
			MPI_Bcast( all_dates, total_datelen, MPI_CHAR, 0, MPI_COMM_WORLD );
			if( debug_msg )	LOG_MSG << "First date: " << all_dates << endl;
		}

		// split into roles
		if( rank == writer_process ){
			writer();
		}else if( rank == scheduler_process ){
			scheduler();
		}else{
			worker();
		}

	}
	catch(const char *error){
		cerr << "Fatal Error:" << endl;
		cerr << error << endl;
		MPI_Abort( MPI_COMM_WORLD, -1 );
		return -1;
	}catch(exception& e ){
		cerr << "Fatal Exception!" << endl;
		cerr << e.what() << endl;
		MPI_Abort( MPI_COMM_WORLD, -1 );
		return -1;
	}
/*	catch(...){
		cerr << "Fatal Unhandled Exception!" << endl;
		MPI_Abort( MPI_COMM_WORLD, -1 );
		
		return -1;
	}
*/	
	// clean up
	cleanupNCBI();

	deleteRegisteredFiles();
	if (remove_db) {
		MPI_Barrier(MPI_COMM_WORLD);
		string local_no_path_sep = config.localPath().substr(0, config.localPath().size() - 1 );
		deleteMpiBlastFiles( config.localPath() );
		if( !removeDirectory( config.localPath() ) )
			perror( "rmdir" );
	}

	return 0;
}

unsigned int determineLastQueryNum(const std::string outputfile){
	string line;
	int lastquerynum = 0;
	streampos lastQpos;
	ifstream outputfile_in(outputfile.c_str());
	while( getline( outputfile_in, line)){
		if ((line.substr(0,9) == "# Query: ") ||
			(line.substr(0,7) == "Query= ") || 
			(line.substr(0,9) == "<?xml ver") ||
			(line.substr(0,13) == "<b>Query=</b>")) {
			lastquerynum++;
			lastQpos = outputfile_in.tellg();
			lastQpos -= outputfile_in.gcount();
		}
	}
	outputfile_in.close();
	ofstream outputfile_out(outputfile.c_str(),ios::app);
	outputfile_out.seekp(lastQpos);
	outputfile_out << "";
	outputfile_out.close();
	
	return(lastquerynum);
}
void makeNewQuery(std::string& new_query, std::string& old_query, unsigned int lastquerynum){
	string line;
	int success = 0;
	int queryN = 0;
	getTempFileName(new_query);
	ifstream qfile_in(old_query.c_str());
	ofstream qfile_out(new_query.c_str());
	while (getline(qfile_in, line)){
		if (line.substr(0,1) == ">")
			queryN++;
		if (queryN > lastquerynum){
			qfile_out << line << endl;
			success++;
		}
	}
	if (success == 0) {
		cerr << "Looks like we don't want to be resuming the run. I think we completed\n";
		terminateProgram(-1);
	}
}

