/******************************************************************************
** mpiblast.hpp
** Interface and variable definitions for mpiBLAST
** $Id: mpiblast.hpp,v 1.33 2005/06/23 12:16:56 koadman Exp $
** $Revision: 1.33 $
** 
** This file is a part of mpiBLAST and is copyright (c)
** by Aaron Darling of Los Alamos National Laboratory for the Regents
** of the University of California.
** 
** License:
** 
** Copyright 2002-2005. The Regents of the University of California.
** 
** This material was produced under U.S. Government contract
** W-7405-ENG-36 for Los Alamos National Laboratory, which is operated
** by the University of California for the U.S. Department of Energy.
** The Government is granted for itself and other acting on its behalf
** a paid-up, non-exclusive, irrevocable worldwide license in the
** material to reproduce, prepare derivative works, and perform
** publicly and display publicly. Beginning five (5) years after
** November 6, 2002, subject to additional five-year worldwide
** renewals, the Government is granted for itself and others acting on
** its behalf a paid-up, nonexclusive, irrevocable worldwide license
** in this material to reproduce, prepare derivative works, distribute
** copies to the public, perform publicly and display publicly, and to
** permit others to do so. NEITHER THE UNITED STATES NOR THE UNITED
** STATES DEPARTMENT OF ENERGY, NOR THE UNIVERSITY OF CALIFORNIA, NOR
** ANY OF THEIR EMPLOYEES, MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR
** ASSUMES ANY LEGAL LIABLITY OR RESPONSIBILITY FOR THE ACCURACY,
** COMPLETENESS, OR USEFULNESS OF ANY INFORMATION, APPARATUS, PRODUCT,
** OR PROCESS DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE
** PRIVATELY OWNED RIGHTS.
** 
** Additionally, this program is free software; you can distribute it
** and/or modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2 of
** the License, or any later version. Accordingly, this program is
** distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or
** FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
** License for more details. You should have received a copy of the
** GNU General Public License along with this program; if not, write
** to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
** Boston, MA 02111-1307, USA.
** 
** ------------------------------------------------------------
** 
** See the file 'COPYING' for the details of the GNU General Public License
** 
******************************************************************************/

#ifndef __mpiblast_hpp__
#define __mpiblast_hpp__

#include "mpi.h"

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include <string>
#include <fstream>
#include <map>
#include <set>
#include <sstream>
#include <vector>
#include <queue>
#include <list>

#ifdef __GNUG__
//#include "unistd.h"
#endif
//#include "stdio.h"

#include "file_util.hpp"
#include "mpiblast_util.hpp"
#include "mpiblast_config.hpp"
#include "mpiblast_types.h"
#include "fragment_list.hpp"
#include "blastjob.hpp"

extern "C" {
#include "blastdef.h"
#include "ncbi.h"
#include "objseq.h"
#include "ncbithr.h"
#include "ncbiwin.h"
#include "connect/ncbi_core_c.h"
}

// Tags for each type of MPI communication
#define	DB_FRAGMENTS_TAG	2	/* A list of database fragments */
#define	BLAST_RETCODE_TAG	7	/* A BLAST return code */
#define	ASSIGNMENT_TYPE		9	/* The type of worker assignment given */
#define	WORKER_IDLE		10	/* That a worker is idle */
#define	FRAGMENT_COPY_COMPLETE	11	/* That a fragment copy has completed */
#define	BLAST_RESULTS_TAG	12	/* Actual BLAST results */
#define	QUERY_TAG		15	/* The query indices that should be searched */
#define DB_DATATOSEND_TAG       17      /* Amount of Data to send when copying DB fragment */
#define DB_EXTENSION_DATA_TAG   18      /* Actual data sent when copying DB fragment */
#define DB_FILESIZE_TAG         19      /* Total amount of data to receive when copying DB fragment */

// assignment types
#define SEARCH_COMPLETE		0 /* Code sent to a worker when the search is complete */ 
#define COPY_FRAGMENT		1 /* Code indicating that the worker should copy a fragment */
#define SEARCH_FRAGMENT		2 /* Code telling a worker to search a fragment */
#define WORKER_QUIT		3 /* Code telling a worker to quit */
#define DO_NOTHING		4 /* place holder telling a worker to do nothing */
#define SEND_RESULTS	5 /* An assignment from a writer to send results */

// process roles:
// jobs are assigned by a scheduler process and results are
// written by a writer process.  The workers need to know which
// rank each of these processes have.
static int writer_process = 0;
static int scheduler_process = 1;

/**
 * This class stores information related to each query--
 * A BlastJob to track how much of the query has completed
 * and the bioseq_cache for bioseqs that go along with the query results
 */
class QueryData {
public:
	QueryData( BlastJob* bj, ValNodePtr bs_cache ) :
		blast_job( bj ), bioseq_cache( bs_cache ),
		searched_frags( 0 ) {};
	QueryData( const QueryData& qd ) :
	blast_job( qd.blast_job ),
	bioseq_cache( qd.bioseq_cache ),
	searched_frags( qd.searched_frags ),
	bioseq_lengths( qd.bioseq_lengths ),
	bioseq_starts( qd.bioseq_starts ),
	bioseq_strands( qd.bioseq_strands )
	{}
	QueryData& operator=( const QueryData& qd ){
		blast_job = qd.blast_job;
		bioseq_cache = qd.bioseq_cache;
		searched_frags = qd.searched_frags;
		bioseq_lengths = qd.bioseq_lengths;
		bioseq_starts = qd.bioseq_starts;
		bioseq_strands = qd.bioseq_strands;
		return *this;
	}

	/** A BlastJob tracks which segments this query has been searched against */
	BlastJob* blast_job;
	int searched_frags;	/**< counts the number of fragments that have been searched */

	// writer only data
	/** 
	 * A cache for any Bioseqs that will be needed during result output.
	 * These get sent from the worker nodes that performed the search.
	 */
	ValNodePtr bioseq_cache;
	std::vector< int > bioseq_lengths;	/**< a list of the original bioseq lengths */
	std::vector< int > bioseq_starts;	/**< a list of the original bioseq starts */
	std::vector< int > bioseq_strands;	/**< a list of the original bioseq strands */
};

typedef struct ResultData 
{
	/** the search complete notification message sent by the worker 
	 *  The message is an array of integers of the form:
	 *  result_id, frag_id, searched count, searched list, success count, success list, data size
	 *  result_id:	a unique identifier the worker has assigned to this result set.
	 *				This ID gets used to request the results from the worker
	 *  frag_id:	The id of the database fragment that was searched
	 *  searched count:	The number of queries searched
	 *  searched list:	The list of searched queries
	 *  success count:	The number of queries that had hits
	 *  success list:	The list of queries with hits
	 *  data size:	The size of the result buffer in bytes
	 */
	int* notification;
	/** A buffer to store result data */
	unsigned char* buf;
	/** The MPI_Request object tracking the progress of the MPI_Irecv() operation to
	 *  receive the results */
	MPI_Request* req;
} ResultData_t;

/**
 * The MpiBlast class provides the functionality to perform BLAST
 * searches on a cluster of nodes connected with MPI.
 */
class MpiBlast{
	public:
		MpiBlast()
		{ 
			log_stream = &std::cerr;  
			remove_db = false;
			output_format = -1;
			copy_via = COPY_VIA_CP;	// default to standard filesystem copy
			disable_mpi_db = false;	// default to workers sending bioseqs
			total_datelen = 0;
			all_dates = NULL;
			concurrent_accesses = 1;
			result_id = 0;
		}
		
		~MpiBlast();
		
		
//
// begin Writer-specific functions
//
		/**
		 * Writer - Reads results in ASN.1 format from a worker
		 */
		void receiveResults( ResultData_t& rd, std::vector<SeqAlignPtr>& results );
		
		/**
		 * Writer - Sends all DB fragment files for "fragment_id" to worker 
		 * "worker_id" using MPI
		 */
		bool sendFragsViaMPI(int fragment_id, int worker_id);
                
		/**
		 * Writer - Free memory used by bioseqs for the given query
		 */
		void purgeBioseqCache( int queryI );
		
		/**
		 * Writer - process received BLAST results
		 * scan through the list of pending result receive operations
		 * when an operation has completed, merge the results with any
		 * previously received results for those queries
		 */
		void processResultReceiveCompletions();
		void readPartialBioseqs( SeqAlignPtr sap_p, AsnIoMemPtr aimp, int queryI );

		/**
		 * Writer - check whether a query has been completely searched, and if so,
		 *          write it using the NCBI toolbox code
		 */
		void outputQueryResults();

		/**
		 * Writer - Restores a partial bioseq to its original size, filling 
		 *			unused portions with junk data.
		 */
		void expandPartialBioseqs( int queryI );
//
// begin Worker-specific functions
//
		/**
		 * Worker -  Receives all DB fragment files for "fragment_id"
		 * from rank 0 using MPI
		 */
		bool recvFragsViaMPI(int fragment_id);

		/**
		 * Worker - fill alias file with the names of database fragments
		 */		
		void WriteAliasFile( const std::string &alias_filename, const std::vector< int > &fragment_list);

		/**
		 * Worker - Tell the writer the search has completed and how 
		 *          large the generated results are
		 * @param	dest	The writer process rank
		 * @param	frag_id	The fragment id that was searched
		 * @param	queries	A vector of query indexes that were searched
		 */
		void notifyWriterOfCompletion( int dest, int frag_id, std::vector< int >& queries );

		/*
		 * quickly check whether the writer wants results
		 */
		void checkResultRequests();

		/**
		 * A process that schedules search jobs
		 */
		void scheduler();
		/**
		 * A process that merges results and writes output
		 */
		void writer();
		/**
		 * A 'worker' process that actually searches the database
		 */
		void worker();

		/**
		 * Call this with the program arguments to start mpiBLAST
		 */
		int main( int argc, char* argv[] );
		
		/** True if the user requested the database be removed */
		bool remove_db;	

	protected:
		MpiBlastConfig config;	/**< Class to read in configuration file */
		FragmentListFile frag_list_file;	/**< Class to read and modify local fragment list file */
		std::string exec_path;		/**< Directory that mpiblast was executed from */
		std::ofstream log_file, pro_phile;	/**< output file streams for logging and profiling information */
		std::string query_filename;	/**< Full path to query file */
		std::string output_filename;	/**< Full path to output file */
		std::string database_name;	/**< Name of database to query */
		std::string blast_cl;		/**< Base BLAST command line including user specified options */
		std::string blast_type;		/**< Type of blast search ( blastn, blastp, etc. ) */
		std::string db_type;			/**< n for nucleotide, p for protein */
		std::string query_file;		/**< Just the filename of the query file (pathname stripped) */
		std::string local_query_filename;	/**< path to a local copy of the query for workers */
		std::string filter_filename;	/**< Full path to gi filter file */
		std::string filter_file;		/**< Just the filename of the gi filter file (pathname stripped) */
		std::string local_filter_filename;	/**< path to a local copy of the gi filter for workers */
		bool disable_mpi_db;	/**< when true, workers don't send bioseqs to writer, instead
									     writer looks them up on disk */

		std::vector< std::string > scheduler_opts;	/**< Command line options for the master node to generate output */
		std::vector< std::string > writer_opts;	/**< Command line options for the master node to generate output */
		std::vector< std::string > worker_opts;	/**< Command line opts for each worker node to search correctly */
		
		int concurrent_accesses;	/**< number of concurrent accesses to shared storage */
		int copy_via;			/**< how to copy the fragments */
		unsigned int total_frags;
		
		// master-specific data structures
		/** Classes to track progress of queries and distribution of database */
		std::vector< QueryData > query_data;
		std::map< int, std::vector< std::pair< int, SeqAlignPtr > > > results_map;
		int output_queryI;	/**< Stores the index of the next query to output */
		std::list< ResultData > result_data;	/**< A list of results that are in the process of being recv'd */
		
		int query_count;	/**< The number of queries in the query file */
		int output_format;	/**< The user-selected output format, defaults to -1 */

		int total_datelen;	/**< size of storage for fragment datestamps */
		char* all_dates;	/**< storage for fragment datestamps */

		// worker-specific data structures
		std::map< int, std::pair< int, unsigned char* > > result_id_data_map; /**< maps result ID's to a pair of buffer size and result buffer */
		int result_id;

private:
		/**
		 * This internal function gets called by the other sendResults()
		 */
		void sendResults( int dest, int queryI);

};


unsigned int determineLastQueryNum(const std::string outputfile);
void makeNewQuery(std::string& queryfile, std::string& query_filename, unsigned int lastquerynum);
unsigned int fragmentCount(const std::string sharedpath, const std::string dbname, const std::string dbtype);
#endif  // __mpiblast_hpp__
