/******************************************************************************
** mpiblast_config.cpp
** Parses the mpiBLAST configuration file
** $Id: mpiblast_config.cpp,v 1.13 2005/12/20 05:26:23 koadman Exp $
** $Revision: 1.13 $
** 
** This file is a part of mpiBLAST and is copyright (c)
** by Aaron Darling of Los Alamos National Laboratory for the Regents
** of the University of California.
** 
** License:
** 
** Copyright 2002-2005. The Regents of the University of California.
** 
** This material was produced under U.S. Government contract
** W-7405-ENG-36 for Los Alamos National Laboratory, which is operated
** by the University of California for the U.S. Department of Energy.
** The Government is granted for itself and other acting on its behalf
** a paid-up, non-exclusive, irrevocable worldwide license in the
** material to reproduce, prepare derivative works, and perform
** publicly and display publicly. Beginning five (5) years after
** November 6, 2002, subject to additional five-year worldwide
** renewals, the Government is granted for itself and others acting on
** its behalf a paid-up, nonexclusive, irrevocable worldwide license
** in this material to reproduce, prepare derivative works, distribute
** copies to the public, perform publicly and display publicly, and to
** permit others to do so. NEITHER THE UNITED STATES NOR THE UNITED
** STATES DEPARTMENT OF ENERGY, NOR THE UNIVERSITY OF CALIFORNIA, NOR
** ANY OF THEIR EMPLOYEES, MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR
** ASSUMES ANY LEGAL LIABLITY OR RESPONSIBILITY FOR THE ACCURACY,
** COMPLETENESS, OR USEFULNESS OF ANY INFORMATION, APPARATUS, PRODUCT,
** OR PROCESS DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE
** PRIVATELY OWNED RIGHTS.
** 
** Additionally, this program is free software; you can distribute it
** and/or modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2 of
** the License, or any later version. Accordingly, this program is
** distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or
** FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
** License for more details. You should have received a copy of the
** GNU General Public License along with this program; if not, write
** to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
** Boston, MA 02111-1307, USA.
** 
** ------------------------------------------------------------
** 
** See the file 'COPYING' for the details of the GNU General Public License
** 
******************************************************************************/

#include "mpiblast_config.hpp"
#include "file_util.hpp"
#include <iostream>

#ifdef WIN32
// for windows mktemp
#include <io.h>
#endif

using namespace std;

extern "C"{
#include "ncbi.h"
//#include "ncbiwin.h"
}


MpiBlastConfig::MpiBlastConfig( const string& filename ) {
	
	if( filename == ".ncbirc" ){
		//
		// try reading the .ncbirc config file for mpiBLAST values
		//
		char path_buffer[ PATH_MAX ] = "";
		char* tmp_path;
	
		// read shared path
		FindPath( "ncbi", "mpiBLAST", "Shared", path_buffer, PATH_MAX );
		shared_db_path = path_buffer;
		// try environment if .ncbirc failed
		if( shared_db_path.length() == 0 && 
			(tmp_path = getenv( "MPIBLAST_SHARED" )) != NULL ){
			shared_db_path = tmp_path;
		}
		// if it still wasn't found then throw an exception!
		if( shared_db_path.length() == 0 )
			throw "Unable to read mpiBLAST shared storage path from either .ncbirc or the MPIBLAST_SHARED environment variable.\nThe mpiBLAST configuration in .ncbirc should look like:\n[mpiBLAST]\nShared=/path/to/shared/storage\nLocal=/path/to/local/storage";
			
		// read local path
		FindPath( "ncbi", "mpiBLAST", "Local", path_buffer, PATH_MAX );
		local_db_path = path_buffer;
		if( local_db_path.length() != 0 && 
			!checkCreateDir( local_db_path ) && 
			(tmp_path = getenv( "MPIBLAST_LOCAL" )) != NULL ){
			cerr << "Warning:  local db path given as " << local_db_path << " in .ncbirc is unreadable.  Trying MPIBLAST_LOCAL environment variable.\n";
			local_db_path = tmp_path;
		}
		if( local_db_path.length() == 0 && 
			(tmp_path = getenv( "MPIBLAST_LOCAL" )) != NULL ){
			local_db_path = tmp_path;
		}

		if( local_db_path.length() != 0 && 
			!checkCreateDir( local_db_path ) && 
			(tmp_path = getenv( "TMPDIR" )) != NULL ){
			cerr << "Warning:  local db path given as " << local_db_path << " in .ncbirc is unreadable.  Trying $TMPDIR/mpiBLAST_local_db.\n";
			local_db_path = tmp_path + PATH_SEPARATOR + "mpiBLAST_local_db";
		}

		// try setting it to $TMPDIR/mpiBLAST_local_db and print a warning
		if( local_db_path.length() == 0 && 
			(tmp_path = getenv( "TMPDIR" )) != NULL ){
			local_db_path = tmp_path + PATH_SEPARATOR + "mpiBLAST_local_db";
			cerr << "Warning: mpiBLAST local storage path unspecified in either .ncbirc or the environment, defaulting to " << local_db_path << endl;
		}
#ifdef WIN32
		// try setting it to $TEMP/mpiBLAST_local_db under windows
		if( local_db_path.length() == 0 && 
			(tmp_path = getenv( "TEMP" )) != NULL ){
			local_db_path = tmp_path + PATH_SEPARATOR + "mpiBLAST_local_db";
			cerr << "Warning: mpiBLAST local storage path unspecified in either .ncbirc or the environment, defaulting to " << local_db_path << endl;
		}
#else
		// finally default to /tmp/mpiBLAST_local_db on unix
		if( local_db_path.length() == 0  ){
			local_db_path = "/tmp/mpiBLAST_local_db";
			cerr << "Warning: mpiBLAST local storage path unspecified in either .ncbirc or the environment, defaulting to " << local_db_path << endl;
		}
#endif		
		// if it still wasn't found then throw an exception!
		if( local_db_path.length() == 0 )
			throw "Unable to read mpiBLAST local storage path from either .ncbirc or the MPIBLAST_LOCAL environment variable.\nThe mpiBLAST configuration in .ncbirc should look like:\n[mpiBLAST]\nShared=/path/to/shared/storage\nLocal=/path/to/local/storage";
		
		// make sure shared and local paths exist
		if( !checkCreateDir( shared_db_path ) ){
			string* error_msg = new string("The shared storage directory " + shared_db_path + " does not exist\n");
			throw error_msg->c_str();
		}

		if( !checkCreateDir( local_db_path ) ){
			string* error_msg = new string("The local storage directory " + local_db_path + " does not exist\n");
			throw error_msg->c_str();
		}
		
		if( !checkDirWritePerms( local_db_path ) ){
			string* error_msg = new string("The local storage directory " + local_db_path + " appears to be write-protected\n");
			throw error_msg->c_str();
		}

		// ensure that the last character is a path separator
		if( shared_db_path.substr( shared_db_path.size() - 1 ) != PATH_SEPARATOR )
			shared_db_path += PATH_SEPARATOR;
		if( local_db_path.substr( local_db_path.size() - 1 ) != PATH_SEPARATOR )
			local_db_path += PATH_SEPARATOR;

		return;
	}

	// Do we need to make a local copy of the config file?
  
  	if( !isRemotePath( filename ) ){
  		// The file is local or accessible via nfs
		config_filename = filename;
  	}
  	else{
  		// Make a copy of the file in the current directory
		// Define the local name by (a) stripping off the hostname and
		// (b) appending a random six character string to the end.
	  	string::size_type pos = filename.find(":", 0);
		pos ++;
		config_filename = filename.substr(pos, filename.length() - pos) + "XXXXXX";
  
  		try{
			getTempFileName(config_filename);
		}
		catch(const char*){
			cerr << "Error creating temp config file:" << endl;
			throw "Error creating temp config file";
		}
	
		copyFile(filename, config_filename, COPY_VIA_XCP);
  	}
  
  
  	// read config file
  	ifstream config_file( config_filename.c_str() );
	
  	if( !config_file.is_open() ){
    		cerr << "Error opening configuration file: " << config_filename << endl;
		cerr << "Try specifying the configuration file location with the --config-file option\n";
    		throw __FILE__ "(MpiBlastConfig): Unable to open config file";
  	}
  
  	// read shared path name
  	getline( config_file, shared_db_path );
	
  	// read local path name
  	getline( config_file, local_db_path );
	
  	// read BLAST binaries path name
  	getline( config_file, blast_path );
	
  	config_file.close();
  
  	// Delete the temporary local file name if we created one
  	if( isRemotePath( filename ) ){
  		unlink( config_filename.c_str() );
  	}

	// ensure that the last character is a path separator
	if( shared_db_path.substr( shared_db_path.size() - 1 ) != PATH_SEPARATOR )
		shared_db_path += PATH_SEPARATOR;
	if( local_db_path.substr( local_db_path.size() - 1 ) != PATH_SEPARATOR )
		local_db_path += PATH_SEPARATOR;
}

MpiBlastConfig::MpiBlastConfig( const MpiBlastConfig& mbc )
{
  *this = mbc;
}

MpiBlastConfig& MpiBlastConfig::operator=( const MpiBlastConfig& mbc )
{
  config_filename = mbc.config_filename;
  local_db_path = mbc.local_db_path;
  shared_db_path = mbc.shared_db_path;
  blast_path = mbc.blast_path;
  return *this;
}


string MpiBlastConfig::defaultConfigFileName()
{
	static string file_name;

// check for definition of the default config file location
#ifdef DEFAULT_CONFIG_FILENAME
	file_name = DEFAULT_CONFIG_FILENAME;
#endif

#ifdef WIN32
	if( file_name == "" ){
		const char* home_path = getenv( "USERPROFILE" );
		if( home_path != NULL ){
			file_name = home_path;
			file_name += PATH_SEPARATOR;
			file_name += ".mpiblastrc";
			if( doesFileExist( file_name ) )
				return file_name;
		}
		// try the windows system directory 
		const char* winnt_path = getenv( "windir" );
		if( winnt_path != NULL ){
			file_name = winnt_path;
			file_name += PATH_SEPARATOR;
			file_name += "mpiblast.ini";
		}else
			file_name = "mpiblast.conf";
	}
#else
	if( file_name == "" ){
		const char* home_path = getenv( "HOME" );
		if( home_path != NULL ){
			file_name = home_path;
			file_name += PATH_SEPARATOR;
			file_name += ".mpiblastrc";
			if( doesFileExist( file_name ) )
				return file_name;
		}
		// use the mpiblast.conf in the installation path
		file_name = INSTALL_PREFIX;
		file_name += PATH_SEPARATOR + "etc" + PATH_SEPARATOR + "mpiblast.conf";
	}
#endif
	return file_name;
}

const string& MpiBlastConfig::TempDirBaseName() {
	static string base_name = "mpiblast_temp";
	return base_name;
}

void MpiBlastConfig::createLocalTempDir() 
{
	string tmpdir_basename = local_db_path + TempDirBaseName() + "XXXXXX";
	char* tmpdirname = new char[ tmpdir_basename.length() + 1 ];
	strncpy( tmpdirname, tmpdir_basename.c_str(), tmpdir_basename.length() + 1 );
#ifndef WIN32
	char* success = mkdtemp( tmpdirname );
#else
	char* success = _mktemp( tmpdirname );
	// windows doesn't actually create the directory...
	if( success != NULL )
		checkCreateDir( tmpdirname );
#endif
	if( success == NULL ){
		perror( "Error creating temp directory" );
		throw tmpdirname;
	}
	local_db_path = tmpdirname + PATH_SEPARATOR;
	delete[] tmpdirname;
}


