/******************************************************************************
** mpiblast_util.cpp
** Provides utility functions used by mpiBLAST.
** $Id: mpiblast_util.cpp,v 1.16 2005/12/20 05:12:46 koadman Exp $
** $Revision: 1.16 $
** 
** This file is a part of mpiBLAST and is copyright (c)
** by Aaron Darling of Los Alamos National Laboratory for the Regents
** of the University of California.
** 
** License:
** 
** Copyright 2002-2005. The Regents of the University of California.
** 
** This material was produced under U.S. Government contract
** W-7405-ENG-36 for Los Alamos National Laboratory, which is operated
** by the University of California for the U.S. Department of Energy.
** The Government is granted for itself and other acting on its behalf
** a paid-up, non-exclusive, irrevocable worldwide license in the
** material to reproduce, prepare derivative works, and perform
** publicly and display publicly. Beginning five (5) years after
** November 6, 2002, subject to additional five-year worldwide
** renewals, the Government is granted for itself and others acting on
** its behalf a paid-up, nonexclusive, irrevocable worldwide license
** in this material to reproduce, prepare derivative works, distribute
** copies to the public, perform publicly and display publicly, and to
** permit others to do so. NEITHER THE UNITED STATES NOR THE UNITED
** STATES DEPARTMENT OF ENERGY, NOR THE UNIVERSITY OF CALIFORNIA, NOR
** ANY OF THEIR EMPLOYEES, MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR
** ASSUMES ANY LEGAL LIABLITY OR RESPONSIBILITY FOR THE ACCURACY,
** COMPLETENESS, OR USEFULNESS OF ANY INFORMATION, APPARATUS, PRODUCT,
** OR PROCESS DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE
** PRIVATELY OWNED RIGHTS.
** 
** Additionally, this program is free software; you can distribute it
** and/or modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2 of
** the License, or any later version. Accordingly, this program is
** distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or
** FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
** License for more details. You should have received a copy of the
** GNU General Public License along with this program; if not, write
** to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
** Boston, MA 02111-1307, USA.
** 
** ------------------------------------------------------------
** 
** See the file 'COPYING' for the details of the GNU General Public License
** 
******************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "mpi.h"
#include "mpi_util.h"

#include "mpiblast_util.hpp"
#include "fragment_list.hpp"
#include "mpiblast_types.h"

extern "C"{
#include "ncbi.h"
#include "objseq.h"
#include "blastdef.h"
#include "blast_hooks.h"
#include "ncbithr.h"
#include "ncbiwin.h"
#include "connect/ncbi_core_c.h"
}

#include <fstream>
#include <iomanip>

using namespace std;

/* Logging for debugging and profiling */
bool debug_msg = false;		/* Default, set to true with --debug */
std::ostream* log_stream; 		/**< Output stream for logging information */
int rank;			/**< Rank of the current MPI process */
int node_count;		/**< Number of MPI processes allocated for the job */
double realstarttime ;		/* Given value immediately on first line of main, before MPI_Init */
double prog_start;		/* Given value after MPI_Init and MPE defs */
double prog_end ;		/* Given value right before MPI_Finalize */


void initNCBI( vector< string >& ncbi_opts ){
	//
	// initialize ncbi library
	//
	size_t ncbi_argc = ncbi_opts.size();
	//char** ncbi_argv = new char*[ ncbi_opts.size() ];
	char** ncbi_argv = NULL;
	
	if( debug_msg ) {
		LOG_MSG << "initializing ncbi ...";
	}
	
	if(ncbi_argc > 0){
		ncbi_argv = (char**)calloc(ncbi_argc, sizeof(char*));
	}
	
	for(uint optI = 0; optI < (uint)ncbi_argc; optI++ ){
		if( debug_msg )
			CONT_LOG_MSG << ncbi_opts[ optI ] << " ";
		//ncbi_argv[ optI ] = new char[ ncbi_opts[ optI ].size() + 1 ];
		ncbi_argv[ optI ] =(char*)calloc((ncbi_opts[ optI ].size() + 1), sizeof(char));
		strcpy( ncbi_argv[ optI ], ncbi_opts[ optI ].c_str() );
	}

	if( debug_msg )
		CONT_LOG_MSG << endl;
		
	Nlm_SetupArguments( ncbi_argc, ncbi_argv );
	
	// Note! Do not clean up the memory allocated in this function!
	// (i.e. ncbi_argv and ncbi_argv[i]). It is now owned by the
	// ncbi library. It is not clear if it needs to be deleted with
	// a call to Nlm_FreeCmdLineArguments(char** argv) in cleanupNCBI().
	// Since this would require saving the pointer, ncbi_argv, for
	// future reference, just forget about it (not a lot of memory
	// is lost anyway ...).
	
#ifdef MSC_VIRT
	if ( !_vheapinit(0, 1250, _VM_ALLSWAP) )
	{
		ErrPost( CTX_NCBIOBJ, 1, "Can't open virtual memory" );
		return 1;
	}
#endif

	/* Initialize connection library's logger, registry and lock */
	CONNECT_Init( 0 );
	
	if( debug_msg ) {
		LOG_MSG << "\n(" << rank << ") done initializing ncbi." << endl;
	}
}


void cleanupNCBI(){
	//
	// cleanup ncbi library
	//
	NlmThreadJoinAll();

	Nlm_FreeConfigStruct();
	ErrSetLogfile( NULL, 0 );
	Nlm_ReleaseAppContext();

#ifdef MSC_VIRT
	_vheapterm();
#endif
	NlmThreadDestroyAll();
}


void addOpt(vector< string >& opt_vector, int opt_character, const char* opt_argument)
{
	char opt_char = opt_character;
	string opt_string = "-";
	opt_string += opt_char;
	opt_vector.push_back( opt_string );
	if( opt_argument != NULL ){
		opt_vector.push_back( opt_argument );
	}
}

void SendIntVec(vector<int> &vec, int dest, int tag)
{
	size_t count = vec.size();
	
	MPI_Send(&count, 1, MPI_INT, dest, tag, MPI_COMM_WORLD);

	if(count > 0){
		int *tmp = (int*)calloc(count, sizeof(int));
		
		if(tmp == NULL){
			throw __FILE__ "SendIntVec: Unable to allocate memory for buffer";
		}
		
		for(size_t i = 0;i < count;i++){
			tmp[i] = vec[i];
		}

		MPI_Send(tmp, count, MPI_INT, dest, 
			tag, MPI_COMM_WORLD);
			
		free(tmp);
	}

}

void RecvIntVec(vector<int> &vec, int src, int tag)
{
	int count;
	
	MPI_Recv(&count, 1, MPI_INT, src, tag, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
	
	vec = vector<int>(count);
	
	if(count > 0){
		int *tmp = (int*)calloc(count, sizeof(int));
		
		if(tmp == NULL){
			throw __FILE__ "RecvIntVec: Unable to allocate memory for buffer";
		}
		
		MPI_Recv(tmp, count, MPI_INT, src, tag, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
		
		for(int i = 0;i < count;i++){
			vec[i] = tmp[i];
		}
			
		free(tmp);
	}
}

void broadcastFile( string& b_filename ){
	ifstream b_file;
	char* b_buf;
	
	b_file.open( b_filename.c_str(), ios::binary );
	
	if( !b_file.is_open() ){
		cerr << "Error opening \"" << b_filename << "\"\n";
		throw __FILE__ "(MpiBlast::broadcastFile): Unable to open file";
	}
	
	b_file.seekg( 0, ios::end );
	
	uint b_filesize = b_file.tellg();
	
	if( debug_msg ){
		LOG_MSG << "broadcasting file size of " << b_filesize << endl;
	}
	
	MPI_Bcast( &b_filesize, 1, MPI_UNSIGNED, 0, MPI_COMM_WORLD );
	
	if( debug_msg ){
		LOG_MSG << "file size broadcasted\n";
	}
	
	//b_buf = new char[ b_filesize ];
	b_buf = (char*)calloc(b_filesize, sizeof(unsigned char));
	
	if(b_buf == NULL){
		throw __FILE__ "(MpiBlast::broadcastFile): Unable to allocate memory";
	}
	
	b_file.seekg( 0, ios::beg );
	
	if( b_file.read( b_buf, b_filesize ) ){
		uint b_read = b_file.gcount();
		if( b_read != b_filesize ){
		
			free(b_buf);
			
			cerr << "Only read " << b_read << " of " << b_filesize << " bytes\n";
			cerr << "Error completely reading \"" << b_filename << "\"\n";
			
			throw __FILE__ "(MpiBlast::broadcastFile): Error completely reading file";
		}
		
		if( debug_msg ){
			LOG_MSG << "broadcasting file\n";
		}
		
		MPI_Bcast( b_buf, b_read, MPI_BYTE, 0, MPI_COMM_WORLD );
		
		if( debug_msg ){
			LOG_MSG << "file broadcasted\n";
		}
	}else{
		free(b_buf);
		
		cerr << "Error reading\"" << b_filename << "\"\n";
		
		throw __FILE__ "(MpiBlast::broadcastFile):Error reading file";
	}
	
	b_file.close();
	
	free(b_buf);
}


void recvBroadcastFile( string& b_filename ){
	ofstream b_file;
	char* b_buf;

	if( debug_msg ){
		LOG_MSG << "waiting for file size broadcast\n";
	}
	
	uint b_filesize;
	
	MPI_Bcast( &b_filesize, 1, MPI_UNSIGNED, 0, MPI_COMM_WORLD );
	
	if( debug_msg ){
		LOG_MSG << "received file size broadcast of " << b_filesize << endl;
	}
	
	//b_buf = new char[ b_filesize ];
	b_buf = (char*)calloc(b_filesize, sizeof(unsigned char));
	
	if(b_buf == NULL){
		throw __FILE__ "(MpiBlast::recvBroadcastFile): Unable to allocate memory";
	}
	
	if( debug_msg ){
		LOG_MSG << "opening receive file " << b_filename << endl;
	}
	
	b_file.open( b_filename.c_str() );

	if(!b_file.is_open()){
		free(b_buf);
		cerr << "Error opening file: " << b_filename.c_str() << endl;
		throw __FILE__ "(MpiBlast::recvBroadcastFile): Unable to open file";
	}
	
	if( debug_msg ){
		LOG_MSG << "receiving file to " << b_filename << endl;
	}
	
	MPI_Bcast( b_buf, b_filesize, MPI_BYTE, 0, MPI_COMM_WORLD );
	
	if( debug_msg ){
		LOG_MSG << "received file broadcast\n";
	}
	
	b_file.write( b_buf, b_filesize );
	b_file.close();
	
	free(b_buf);
}


uint64 	max_data_to_send = 1024*1024;	/**< max data to send when using MPI to copy fragments */

void sendFile( const string& filename, int dest, int tag ){
	ifstream file_input( filename.c_str() );
	if( !file_input.is_open() ){
		LOG_MSG << "Unable to open \"" << filename << "\"\n";
		throw "Unable to open file for send.";
	}
	
	// send the file size
	file_input.seekg( 0, ios::end );
	uint filesize = file_input.tellg();	
	MPI_Send( &filesize, 1, MPI_UNSIGNED, dest, tag, MPI_COMM_WORLD );
	file_input.seekg( 0, ios::beg );

	// read and send the file
	char* buffer;
	int bufsize = max_data_to_send;
	meta_MPI_Alloc_mem( bufsize, MPI_INFO_NULL, &buffer );

	while( true ){
		file_input.read( buffer, bufsize );
		int count = file_input.gcount();
		if( count <= 0 )
			break;
		MPI_Send( buffer, count, MPI_BYTE, dest, tag, MPI_COMM_WORLD );
	}
	meta_MPI_Free_mem( buffer );
}

void recvFile( const string& filename, int src, int tag ){
	ofstream output_file;
	char* buffer;
	int bufsize = max_data_to_send;
	
	output_file.open( filename.c_str() );
	if( !output_file.is_open() ){
		throw "Unable to open file";
	}

	uint b_filesize;
	
	MPI_Recv( &b_filesize, 1, MPI_UNSIGNED, src, tag, MPI_COMM_WORLD, MPI_STATUS_IGNORE );
	
	meta_MPI_Alloc_mem( bufsize, MPI_INFO_NULL, &buffer );
	
	while( output_file.tellp() < b_filesize ){
		int remaining = b_filesize - output_file.tellp() < bufsize ? b_filesize - output_file.tellp() : bufsize;
		MPI_Recv( buffer, remaining, MPI_BYTE, src, tag, MPI_COMM_WORLD, MPI_STATUS_IGNORE );
		output_file.write( buffer, remaining );
	}
	output_file.close();
	meta_MPI_Free_mem( buffer );
}

void printProgress (uint64 progress_sofar, float64 total) {
	int temp_ratio = (int)((progress_sofar/total)*10000);
	float ratio = (float)temp_ratio / 100.0 ;
	cout << setprecision(2);
	cout << "\b\b\b\b\b" << ratio << "%" ;
	if (progress_sofar == total) 
		cout << "\b\b\b\b\b\b" ;
	cout.flush();
}

//Lock or unlock an entire file
#ifndef WIN32
int getlock (int fd, int cmd, int type) {
	struct flock lock;
	lock.l_type = type;  //F_RDLCK , F_WRLCK, F_UNLCK
	lock.l_start = 0;
	lock.l_whence = SEEK_SET;
	lock.l_len = 0;
	// only lock the file if locking was enabled--
	// some NFS systems don't behave well with locking
	// enabled so we only want to use it when absolutely
	// necessary
	if( lock_mbf )
		return( fcntl(fd,cmd,&lock));
	else
		return 0;
}
#endif
