#!/usr/bin/perl -w
use strict;
use Env qw(HOME TMPDIR MPIBLAST_LOCAL MPIBLAST_SHARED PATH);
use Cwd;


my $mpiblast_dir = getcwd();
my $mpiformatdb = "$mpiblast_dir/src/mpiformatdb";
my $mpiblast = "$mpiblast_dir/src/mpiblast";
my $LocalDir = getconfig('Local');
my $SharedDir = getconfig('Shared');
my $DataDir = getconfig('Data');
$DataDir =~ /(.*\/)data/;
my $NCBI_BinDir = "$1bin";
my $blastall = "$NCBI_BinDir/blastall";
my $formatdb = "$NCBI_BinDir/formatdb";
my $cpus = countcpus();
my $mpirun = "mpirun -nolocal -machinefile /home/lcarey/machines -np 3";

my $too_small = 30; #set to -1 to allow 0-legnth seqs to be created

#print "Local, Shared, Data: $LocalDir, $SharedDir, $DataDir\n";

#Do we want to check mpiformatdb or mpiBLAST?

my $what = shift;
unless ($what){
	print "SYNTAX: $0 [mpiformatdb|mpiblast|ncbi-toolkit|make-database|automake|check_everything|resume]\n";
	exit(0);
}
check_mpiformatdb(@ARGV) if ($what =~ /mpiformatdb/);
check_mpiblast(@ARGV) if ($what =~ /mpiblast/);
make_database(@ARGV) if ($what =~ /make-database/);
check_toolkit(@ARGV) if ($what =~ /ncbi-toolkit/);
check_all(@ARGV) if ($what =~ /check_everything/);
automake(@ARGV) if ($what =~ /automake/);
check_resume(@ARGV) if ($what =~ /resume/);
print "SYNTAX: $0 [mpiformatdb|mpiblast|ncbi-toolkit|make-database|automake|resume]\n";
exit(0);





#check mpiformatdb: MaxN [increment] database1 [database2] ... [databaseN] 
sub check_mpiformatdb {
	my $syntax = "SYNTAX: $0 mpiformatdb MaxNumberFrags [increment] database1 [database2] ... [databaseN] \n";
	my $MaxN = shift;
	die $syntax unless ($MaxN =~ /^[0-9]+$/);
	my $increment = 1;
	$increment = shift if ($_[0] =~ /^[0-9]+$/);
	foreach my $db (@_) {
		$db =~ m/.*\/(\w+\.*\w+)/;
		my $db_fname = $1 || $db;
		unless (-r $db) {
			print "database $db is unreadable. Make sure the file exits\n";
			next;
		}
		for(my $N = 1 ; $N <= $MaxN ; $N += $increment){
			`rm -f $SharedDir/$db_fname.??? $SharedDir/$db_fname.???.???`;
			my $cmd = "$mpiformatdb -N $N -i $db";
			print "Running $cmd\n";
			my $exstat = (system("$cmd") >> 8);
			$increment = ($MaxN - $N) if ( ($increment + $N > $MaxN) && ($N != $MaxN) ); 
			die "$cmd has exited with exit status: $exstat\n" unless ($exstat == 0);
			my $files_created = `ls $SharedDir/$db_fname.??? $SharedDir/$db_fname.???.???  | wc -l`;
			$files_created =~ s/\s*(\d+)\s*\n/$1/;
			my $should_have = $N*7+2;
			die "$cmd \n\tproduced $files_created instead of $should_have files\n" unless ($files_created == $should_have);
		}
	}
	exit(0);
}


sub log10 {
	my $n = shift;
	return log($n)/log(10);
}

#"SYNTAX: $0 make-database dabase_name.[aa|nt] db_size(in Millions of letters)";
#db_size can be a decimal, to create tiny databases
sub make_database {
	my @Letters;
	my @SeqLengthArray;
	my $MAXSIZE=4; #Max order of magnitude of size of a seq
	my $syntax = "SYNTAX: $0 make-database dabase_name.[aa|nt] db_size(in Millions of letters)";
	my $db = shift;
	my $size = (1000000 * shift) || die $syntax;
	die $syntax unless ( ($size =~ /^\d+$/) && ( $db =~ /^([\w\/]+\.)+(nt|aa)$/) ) ;
	#die "DB file \'$db\' already exists\n" if (-e $db);
	@Letters=qw(A C D E F G H I K L M P Q S T V Y) if ($2 eq 'aa');
	@Letters=qw(A C T G) if ($2 eq 'nt');

	while($size>0){
		my $order_mag = (int(rand($MAXSIZE))+1);
		my $n = int(rand(1*10**$order_mag));
			if (($n < $size+50) && ($n > $too_small)){
				push(@SeqLengthArray, $n) ;
				$size-=$n;
			}
	}

	open(DB,">$db");
	my $N = 0;
	foreach (@SeqLengthArray) {
	#	print DB ">gi|$N|$db Randomly Generated seq# $N of size $_\n";
		print DB ">${N}_$db\n";
		for(my $L=1;$L<=($_);$L++){
			print DB  $Letters[ rand @Letters ];
		}
		$N++;
		print DB "\n";
	}
	close(DB);

}


sub check_toolkit {
	my @months=qw(Jan Feb Mar Apr May Jun Jly Aug Sep Oct Nov Dec);
	my $ncbi_dir = shift;
	die "SYNTAX: $0 ncbi-toolkit /path/to/ncbi\n" unless (-d $ncbi_dir);
	my $rmcmd = "rm -f ${ncbi_dir}/build/*.o ${ncbi_dir}/lib/* ";
	chdir "$ncbi_dir/../";
	my $keyfile = "ncbi/doc";
	my $modtime = (stat($keyfile))[9];
	my $modmonth = ((localtime($modtime))[4]+1);
	my $modyear = ((localtime($modtime))[5]-100+2000);
	my $Abbr = $months[$modmonth -1];
	print "$keyfile last modified on $modmonth/${modyear}. We can patch this with ncbi_$Abbr$modyear\*\.patch.\n";
	my $patched = system("grep -q mpiBLAST $keyfile");
	if (($patched >> 8) == 1) { #need to patch
		print "Patching the NCBI toolkit located at $ncbi_dir\n";
		system("patch -p0 < $mpiblast_dir/ncbi_$Abbr${modyear}_evalue.patch");
		print "Would you like me cleanout the ncbi-toolkit and rebuild it?\n";
		print "I will run: \'$rmcmd\'\n";
		print "If not, you will have to do this yourself before building mpiBLAST.  (y/N)?\n";
		my $answer = getc();
		if ($answer eq "Y" || $answer eq "y") {
			`$rmcmd`;
			`ncbi/make/makedis.csh`;
		}
		exit(0);
	}
	exit(0);
}

sub getconfig {
	my $what = $_[0];
	my $LocalDir = $TMPDIR;
	my $SharedDir;
	my $DataDir;

	open(CONFIG, "$HOME/.ncbirc");
	while(<CONFIG>){
		$DataDir = $1 if (/^Data=(.*)/);
		$SharedDir = $1 if (/^Shared=(.*)/);
		$LocalDir = $1 if (/^Local=(.*)/);
	}
	close(CONFIG);

	$LocalDir = $MPIBLAST_LOCAL if ($MPIBLAST_LOCAL);
	$SharedDir = $MPIBLAST_SHARED if ($MPIBLAST_SHARED);

	unless ($LocalDir && $SharedDir && $DataDir ) {
		print "Unable to get all required variables from either the environment or ~/.ncbirc.\n";
		print "Please read the README file for information on setting up mpiBLAST\n";
		print "Found the following:\n";
		print "LocalStorage=$LocalDir\n";
		print "SharedStorage=$SharedDir\n";
		print "NCBI Data=$DataDir\n";
		die;
	}
	return $LocalDir if ($what eq 'Local');
	return $SharedDir if ($what eq 'Shared');
	return $DataDir if ($what eq 'Data');
}

#Automatically build the configure script from the files availible on cvs
#Uses the most recent availible versions of autotools
sub automake {
	my $aclocal = 'aclocal';
	my $automake = 'automake';
	my $autoheader = 'autoheader';
	my $autoconf = 'autoconf';
	my @versions=('-1.9','-1.8','-1.7','-1.6');
	foreach my $path ( split(/:/,$PATH) ) {
		foreach my $ver (@versions) {
			my $filename = $path . '/' . 'aclocal' . $ver ;
			if (-f $filename ){
				print "Building configure script using autotools$ver\n";
				system("${aclocal}${ver}"); 
				system("${autoheader}");
				system("${automake}${ver} -a");
				system("${autoconf}");
				exit(0);
			}
		}
	}
}
	

sub check_mpiblast {
	my $database = shift;
	my $query = shift;
	my $type = shift;
	die "SYNTAX: $0 mpiblast database query search_type\n" unless ($type);
	die "Unable to read $query\n" unless (-r $query);
	die "Unable to run $blastall\n" unless (-r $blastall);
	die "Unable to read $SharedDir/$database\n" unless (-r "$SharedDir/${database}.mbf");
	die "blast_type must be one of: blastn, blastp, tbast, blastx, tblastx\n" unless ( $type =~ m/^t{0,1}blast[pnx]{0,1}$/);

	my $outfile = "${query}_vs_${database}-blastall";
	$outfile =~ s/(.*\/)(\w.*)/$2/;

	my $cmd ="$blastall -p $type -d $SharedDir/$database -m 8 -o $outfile -i $query -a $cpus";
	print "First running blastall to check mpiBLAST's results\n $cmd\n";
	system("$cmd");
	$outfile =~ s/(.*-)blastall/$1mpiblast/;
	$cmd = "mpirun -np 3 src/mpiblast -m 8 -o $outfile -i $query -d $database -p $type --debug";
	print "Now running mpiblast with -np 3\n $cmd\n";
	system("$cmd");
	exit(0);
}


sub countcpus {
	return 1 unless (-r "/proc/cpuinfo");
	my $ncpus = 0;
	open(CPU,"/proc/cpuinfo");
	while(<CPU>){
		$ncpus++ if (/^vendor_id/);
	}
	return $ncpus;
}

	
sub check_all {

	print "Makeing DBs\n";
print 	`$0 make-database query.aa .01`;
print 	`$0 make-database db.aa 2`;
print 	`$0 make-database query.nt .01`;
print 	`$0 make-database db.nt 2`;
	
	print "formating DBs\n";
print 	`$0 mpiformatdb 2 db.aa`;
print 	`$0 mpiformatdb 2 db.nt`;
	
	print "Running Blast -- go get some coffee\n";
print 	`$0 mpiblast db.aa query.aa blastp`; 
print 	`$0 mpiblast db.nt query.nt blastn`; 
#print 	`$0 mpiblast db.aa query.aa tblast`; 
#print 	`$0 mpiblast db.aa query.nt blastx`; 
#print 	`$0 mpiblast db.nt query.aa tblastn`; 
print 	`$0 mpiblast db.nt query.nt tblastx`; 

exit(0);
}

sub check_resume {
	make_database('query.aa', .01);
	make_database('query2.aa', .02);
	my $Nq=0;
	my $Nq2=0;
	open(NQ, "grep \\> query.aa | wc -l |");
	read(NQ,$Nq, 15);
	chomp $Nq;
	close(NQ);
	open(NQ, "grep \\> query2.aa | wc -l |");
	read(NQ,$Nq2, 15);
	chomp $Nq2;
	close(NQ);
	print make_database('db.aa', 2);
	print `$mpiformatdb -i db.aa -N 1`;
	print "Testing resume w/-m 9 output\n";
	print "$mpirun $mpiblast -d db.aa -i query.aa -o output -m 9 -p blastp\n";
	print `$mpirun $mpiblast -d db.aa -i query.aa -o output -m 9 -p blastp`;
	`cat 'query2.aa' >> 'query.aa'`;
	print `$mpirun $mpiblast -d db.aa -i query.aa -o output -m 9 -p blastp --resume`;
	my $noutputs = 0;
	open(NQ, "egrep 'Query:|Query=|xml|seq-annot' output | wc -l |");
	read(NQ,$noutputs,15);
	chomp $noutputs;
	close(NQ);
	unless ($Nq+$Nq2 == $noutputs){
		print "$Nq $Nq2 $noutputs ",$Nq+$Nq2,"\n";
		print "--resume seems to be broken on -m 9\n";
		exit(1);
	}

	print "Testing resume w/-m 1 output\n";
	make_database('query.aa', .01);
	make_database('query2.aa', .02);
	print "$mpirun $mpiblast -d db.aa -i query.aa -o output -m 1 -p blastp\n";
	print `$mpirun $mpiblast -d db.aa -i query.aa -o output -m 1 -p blastp`;
	`cat 'query2.aa' >> 'query.aa'`;
	print `$mpirun $mpiblast -d db.aa -i query.aa -o output -m 1 -p blastp --resume`;
	unless ($Nq+$Nq2 == $noutputs){
		print "$Nq $Nq2 $noutputs ",$Nq+$Nq2,"\n";
		print "--resume seems to be broken on -m 9\n";
		exit(1);
	}
	exit(0);
}


